import numpy as np
from scipy.linalg import eigh
from time import time
import pynof

def compute_energy(mol,p=None,C=None,n=None,fmiug0=None,hfidr=True,nofmp2=False,mbpt=False,gradients=False,printmode=True,ekt=False,mulliken_pop=False,lowdin_pop=False,m_diagnostic=False,perturb=False,erpa=False):
 
    t1 = time()

    wfn = p.wfn
    S,T,V,H,I,b_mnl,Dipole = pynof.compute_integrals(wfn,mol,p)

    if(printmode):
        print("Number of basis functions                   (NBF)    =",p.nbf)
        if(p.RI):
            print("Number of auxiliary basis functions         (NBFAUX) =",p.nbfaux)
        print("Inactive Doubly occupied orbitals up to     (NO1)    =",p.no1)
        print("No. considered Strongly Doubly occupied MOs (NDOC)   =",p.ndoc)
        print("No. considered Strongly Singly occupied MOs (NSOC)   =",p.nsoc)
        print("No. of Weakly occ. per St. Doubly occ.  MOs (NCWO)   =",p.ncwo)
        print("Dimension of the Nat. Orb. subspace         (NBF5)   =",p.nbf5)
        print("No. of electrons                                     =",p.ne)
        print("Multiplicity                                         =",p.mul)
        print("")

    # Energía Nuclear
    E_nuc = mol.nuclear_repulsion_energy()

    # Guess de MO (C)
    Cguess = C
    if(C is None):
        E_i,Cguess = eigh(H, S)  # (HC = SCe)
    Cguess = pynof.check_ortho(Cguess,S,p)

    if (hfidr):
        EHF,Cguess,fmiug0guess = pynof.hfidr(Cguess,H,I,b_mnl,E_nuc,p,printmode)

    if(C is None):
        C = Cguess
    C = pynof.check_ortho(C,S,p)

    if(n is None):
        gamma = pynof.compute_gammas_trigonometric(p.nv,p.ndoc,p.ncwo)
        if p.occ_method == "Softmax":
            p.nv = p.nbf5 - p.no1 - p.nsoc 
            n,dn_dgamma = pynof.ocupacion_trigonometric(gamma,p.no1,p.ndoc,p.nalpha,p.nv,p.nbf5,p.ndns,p.ncwo,p.HighSpin)
            gamma = pynof.n_to_gammas_softmax(n,p.nv,p.no1,p.ndoc,p.ndns,p.ncwo)
    else:
        if p.occ_method == "Trigonometric":
            gamma = pynof.n_to_gammas_trigonometric(n,p.nv,p.no1,p.ndoc,p.ndns,p.ncwo)
        if p.occ_method == "Softmax":
            p.nv = p.nbf5 - p.no1 - p.nsoc 
            gamma = pynof.n_to_gammas_softmax(n,p.nv,p.no1,p.ndoc,p.ndns,p.ncwo)

    elag = np.zeros((p.nbf,p.nbf))

    E_occ,nit_occ,success_occ,gamma,n,cj12,ck12 = pynof.occoptr(gamma,C,H,I,b_mnl,p)

    iloop = 0
    itlim = 0
    E,E_old,E_diff = 9999,9999,9999
    Estored,Cstored,gammastored = 0,0,0
    last_iter = 0

    if(printmode):
        print("")
        print("PNOF{} Calculation ({}/{} Optimization)".format(p.ipnof,p.orb_method,p.occ_method))
        print("==================")
        print("")
        print('{:^7} {:^7}  {:^7}  {:^14} {:^14} {:^14}   {:^6}   {:^6} {:^6} {:^6}'.format("Nitext","Nit_orb","Nit_occ","Eelec","Etot","Ediff","Grad_orb","Grad_occ","Conv Orb","Conv Occ"))
    for i_ext in range(p.maxit):
        #orboptr
        #t1 = time()
        if(p.orb_method=="ID"):
            E_orb,C,nit_orb,success_orb,itlim,fmiug0 = pynof.orboptr(C,n,H,I,b_mnl,cj12,ck12,i_ext,itlim,fmiug0,p,printmode)
        if(p.orb_method=="Rotations"):
            E_orb,C,nit_orb,success_orb = pynof.orbopt_rotations(gamma,C,H,I,b_mnl,p)
        #t2 = time()

        #occopt
        E_occ,nit_occ,success_occ,gamma,n,cj12,ck12 = pynof.occoptr(gamma,C,H,I,b_mnl,p)
        #t3 = time()
        #print("t_orb: {:3.1e} t_occ: {:3.1e}".format(t2-t1,t3-t2))
        if(p.occ_method=="Softmax"):
            C,gamma = pynof.order_occupations_softmax(C,gamma,H,I,b_mnl,p)

        E = E_orb
        E_diff = E-E_old
        E_old = E

        y = np.zeros((p.nvar))
        grad_orb = pynof.calcorbg(y,n,cj12,ck12,C,H,I,b_mnl,p)
        J_MO,K_MO,H_core = pynof.computeJKH_MO(C,H,I,b_mnl,p)
        grad_occ = pynof.calcoccg(gamma,J_MO,K_MO,H_core,p)
        print("{:6d} {:6d} {:6d}   {:14.8f} {:14.8f} {:15.8f}      {:3.1e}    {:3.1e}   {}   {}".format(i_ext,nit_orb,nit_occ,E,E+E_nuc,E_diff,np.linalg.norm(grad_orb),np.linalg.norm(grad_occ),success_orb,success_occ))

        if(success_orb or (np.linalg.norm(grad_orb) < 1e-3 and np.linalg.norm(grad_occ)< 1e-3)):

            if perturb and E - Estored < -1e-4:
                y = np.zeros((p.nvar))
                grad_orb = pynof.calcorbg(y,n,cj12,ck12,C,H,I,b_mnl,p)
                J_MO,K_MO,H_core = pynof.computeJKH_MO(C,H,I,b_mnl,p)
                grad_occ = pynof.calcoccg(gamma,J_MO,K_MO,H_core,p)
                print("Increasing Gradient")
                last_iter = i_ext
                Estored,Cstored,gammastored = E,C.copy(),gamma.copy()
                C,gamma = pynof.perturb_solution(C,gamma,grad_orb,grad_occ,p)
            else:
                print("Solution does not improve anymore")
                if(Estored<E):
                    E,C,gamma = Estored,Cstored,gammastored
                break

    if(p.orb_method=="ID"):
        np.save(p.title+"_fmiug0.npy",fmiug0)
    
    n,dR = pynof.ocupacion(gamma,p.no1,p.ndoc,p.nalpha,p.nv,p.nbf5,p.ndns,p.ncwo,p.HighSpin,p.occ_method)
    cj12,ck12 = pynof.PNOFi_selector(n,p)
    E,elag,sumdiff,maxdiff = pynof.ENERGY1r(C,n,H,I,b_mnl,cj12,ck12,p)
    print("\nLagrage sumdiff {:3.1e} maxfdiff {:3.1e}".format(sumdiff,maxdiff))

    C,n,elag = pynof.order_subspaces(C,n,elag,H,I,b_mnl,p)

    np.save(p.title+"_C.npy",C)
    np.save(p.title+"_n.npy",n)

    if(printmode):
        print("")
        print("RESULTS OF THE OCCUPATION OPTIMIZATION")
        print("========================================")

        e_val = elag[np.diag_indices(p.nbf5)]
        print(" {:^3}    {:^9}   {:>12}".format("Idx","n","E (Hartree)"))
        for i in range(p.nbeta):
            print(" {:3d}    {:9.7f}  {:12.8f}".format(i+1,2*n[i],e_val[i]))
        for i in range(p.nbeta,p.nalpha):
            if(not p.HighSpin):
                print(" {:3d}    {:9.7f}  {:12.8f}".format(i+1,2*n[i],e_val[i]))
            else:
                print(" {:3d}    {:9.7f}  {:12.8f}".format(i+1,n[i],e_val[i]))
        for i in range(p.nalpha,p.nbf5):
            print(" {:3d}    {:9.7f}  {:12.8f}".format(i+1,2*n[i],e_val[i]))

        print("")

        print("----------------")
        print(" Final Energies ")
        print("----------------")
        
        if(hfidr):
            print("       HF Total Energy = {:15.7f}".format(E_nuc + EHF))
        print("Final NOF Total Energy = {:15.7f}".format(E_nuc + E))
        if(hfidr):
            print("    Correlation Energy = {:15.7f}".format(E-EHF))
        print("")
        print("")

    E_t = E_nuc + E

    pynof.fchk(p.title,wfn,mol,"Energy",E_t,elag,n,C,p)

    t2 = time()
    print("Elapsed Time: {:10.2f} (Seconds)".format(t2-t1))

    if(nofmp2):
        pynof.nofmp2(n,C,H,I,b_mnl,E_nuc,p)

    if(mbpt):
        pynof.mbpt(n,C,H,I,b_mnl,Dipole,E_nuc,E,p)

    if(ekt):
        pynof.ext_koopmans(p,elag,n)

    if(mulliken_pop):
        pynof.mulliken_pop(p,wfn,n,C,S)

    if(lowdin_pop):
        pynof.lowdin_pop(p,wfn,n,C,S)

    if(m_diagnostic):
        pynof.M_diagnostic(p,n)

    if(erpa):
        pynof.ERPA(wfn,mol,n,C,H,I,b_mnl,cj12,ck12,elag,p)


    if gradients:
        grad = pynof.compute_geom_gradients(wfn,mol,n,C,cj12,ck12,elag,p)
        return E_t,C,n,fmiug0,grad.flatten()
    else:
        return E_t,C,n,fmiug0


def brute_force_energy(mol,p,intents=5,C=None,gamma=None,fmiug0=None,hfidr=True,RI_last=False,gpu_last=False,ekt=False,mulliken_pop=False,lowdin_pop=False,m_diagnostic=False):
    t1 = time()
    
    E,C,gamma,fmiug0 = pynof.compute_energy(mol,p,C,gamma,fmiug0,hfidr)
    E_min = E
    C_min = C
    gamma_min = gamma
    fmiug0_min = fmiug0
    
    for i in range(intents):
        p.autozeros()
        E,C,gamma,fmiug0 = pynof.compute_energy(mol,p,C,gamma,None,hfidr=False,nofmp2=False)
        if(E<E_min):
            E_min = E
            C_min = C
            gamma_min = gamma
            fmiug0_min = fmiug0
    
    p.RI = RI_last
    p.gpu = gpu_last
    p.jit = True
    E,C,gamma,fmiug0 = pynof.compute_energy(mol,p,C_min,gamma_min,fmiug0_min,hfidr=False,nofmp2=False,ekt=ekt,mulliken_pop=mulliken_pop,lowdin_pop=lowdin_pop,m_diagnostic=m_diagnostic)
    
    t2 = time()
    
    print("Best Total NOF Energy {}".format(E))
    print("Elapsed Time: {:10.2f} (Seconds)".format(t2-t1))
   
    return E,C,gamma,fmiug0
    



import numpy as np
import pynof
from numba import prange,njit,jit
from time import time
from scipy.linalg import eigh,expm,solve
from scipy.optimize import root
try:
    import cupy as cp
except:
    pass

def computeF_RC_driver(J,K,n,H,cj12,ck12,p):

    if(p.gpu):
        F = computeF_RC_GPU(J,K,n,H,cj12,ck12,p)
    else:
        F = computeF_RC_CPU(J,K,n,H,cj12,ck12,p)

    return F

def computeF_RC_CPU(J,K,n,H,cj12,ck12,p):

    # Matriz de Fock Generalizada
    F = np.zeros((p.nbf5,p.nbf,p.nbf))

    ini = 0
    if(p.no1>1):
        ini = p.no1

    # nH
    F += np.einsum('i,mn->imn',n,H,optimize=True)        # i = [1,nbf5]

    # nJ
    F[ini:p.nbeta,:,:] += np.einsum('i,imn->imn',n[ini:p.nbeta],J[ini:p.nbeta,:,:],optimize=True)        # i = [ini,nbeta]
    F[p.nalpha:p.nbf5,:,:] += np.einsum('i,imn->imn',n[p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:],optimize=True)  # i = [nalpha,nbf5]

    # C^J J
    np.fill_diagonal(cj12[ini:,ini:],0) # Remove diag.
    F += np.einsum('ij,jmn->imn',cj12,J,optimize=True)                                                # i = [1,nbf5]
    #F[ini:p.nbf5,:,:] -= np.einsum('ii,imn->imn',cj12[ini:p.nbf5,ini:p.nbf5],J[ini:p.nbf5,:,:],optimize=True) # quita i==j

    # -C^K K
    np.fill_diagonal(ck12[ini:,ini:],0) # Remove diag.
    F -= np.einsum('ij,jmn->imn',ck12,K,optimize=True)                                                # i = [1,nbf5]
    #F[ini:p.nbf5,:,:] += np.einsum('ii,imn->imn',ck12[ini:p.nbf5,ini:p.nbf5],K[ini:p.nbf5,:,:],optimize=True) # quita i==j

    return F

def computeF_RC_GPU(J,K,n,H,cj12,ck12,p):

    # Matriz de Fock Generalizada
    F = cp.zeros((p.nbf5,p.nbf,p.nbf))

    ini = 0
    if(p.no1>1):
        ini = p.no1

    # nH
    F += cp.einsum('i,mn->imn',n,H,optimize=True)        # i = [1,nbf5]

    # nJ
    F[ini:p.nbeta,:,:] += cp.einsum('i,imn->imn',n[ini:p.nbeta],J[ini:p.nbeta,:,:],optimize=True)
    F[p.nalpha:p.nbf5,:,:] += cp.einsum('i,imn->imn',n[p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:],optimize=True)

    # C^J J
    np.fill_diagonal(cj12[ini:,ini:],0) # Remove diag.
    F += cp.einsum('ij,jmn->imn',cj12,J,optimize=True)

    # -C^K K
    np.fill_diagonal(ck12[ini:,ini:],0) # Remove diag.
    F -= cp.einsum('ij,jmn->imn',ck12,K,optimize=True)

    return F.get()

def computeF_RO_driver(J,K,n,H,cj12,ck12,p):

    if(p.gpu):
        F = computeF_RO_GPU(J,K,n,H,cj12,ck12,p)
    else:
        F = computeF_RO_CPU(J,K,n,H,cj12,ck12,p)

    return F


def computeF_RO_CPU(J,K,n,H,cj12,ck12,p):

    # Matriz de Fock Generalizada
    F = np.zeros((p.nbf5,p.nbf,p.nbf))

    ini = 0
    if(p.no1>1):
        ini = p.no1

    # nH
    F[:p.nbeta,:,:] += np.einsum('i,mn->imn',n[:p.nbeta],H,optimize=True)                      # i = [1,nbf5]
    F[p.nbeta:p.nalpha,:,:] += 0.5*H                                                           # i = [nbeta,nalpha]
    F[p.nalpha:p.nbf5,:,:] += np.einsum('i,mn->imn',n[p.nalpha:p.nbf5],H,optimize=True)        # i = [nalpha,nbf5]

    # nJ
    F[ini:p.nbeta,:,:] += np.einsum('i,imn->imn',n[ini:p.nbeta],J[ini:p.nbeta,:,:],optimize=True)        # i = [ini,nbeta]
    F[p.nalpha:p.nbf5,:,:] += np.einsum('i,imn->imn',n[p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:],optimize=True)  # i = [nalpha,nbf5]

    # C^J J
    np.fill_diagonal(cj12[ini:,ini:],0) # Remove diag.
    F[:p.nbeta,:,:] += np.einsum('ij,jmn->imn',cj12[:p.nbeta,:p.nbeta],J[:p.nbeta,:,:],optimize=True)                               # i = [1,nbeta]
    F[:p.nbeta,:,:] += np.einsum('ij,jmn->imn',cj12[:p.nbeta,p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:],optimize=True)                               # i = [1,nbeta]
    F[p.nalpha:p.nbf5,:,:] += np.einsum('ij,jmn->imn',cj12[p.nalpha:p.nbf5,:p.nbeta],J[:p.nbeta,:,:],optimize=True)                                      # i = [nalpha,nbf5]
    F[p.nalpha:p.nbf5,:,:] += np.einsum('ij,jmn->imn',cj12[p.nalpha:p.nbf5,p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:],optimize=True)                                      # i = [nalpha,nbf5]
    #F[ini:p.nbf5,:,:] -= np.einsum('ii,imn->imn',cj12[ini:p.nbf5,ini:p.nbf5],J[ini:p.nbf5,:,:],optimize=True) # quita i==j

    # -C^K K
    np.fill_diagonal(ck12[ini:,ini:],0) # Remove diag.
    F[:p.nbeta,:,:] -= np.einsum('ij,jmn->imn',ck12[:p.nbeta,:p.nbeta],K[:p.nbeta,:,:],optimize=True)                                                # i = [1,nbeta]
    F[:p.nbeta,:,:] -= np.einsum('ij,jmn->imn',ck12[:p.nbeta,p.nalpha:p.nbf5],K[p.nalpha:p.nbf5,:,:],optimize=True)                                                # i = [1,nbeta]
    F[p.nalpha:p.nbf5,:,:] -= np.einsum('ij,jmn->imn',ck12[p.nalpha:p.nbf5,:p.nbeta],K[:p.nbeta,:,:],optimize=True)                                      # i = [nalpha,nbf5]
    F[p.nalpha:p.nbf5,:,:] -= np.einsum('ij,jmn->imn',ck12[p.nalpha:p.nbf5,p.nalpha:p.nbf5],K[p.nalpha:p.nbf5,:,:],optimize=True)                                      # i = [nalpha,nbf5]
    #F[ini:p.nbf5,:,:] += np.einsum('ii,imn->imn',ck12[ini:p.nbf5,ini:p.nbf5],K[ini:p.nbf5,:,:],optimize=True) # quita i==j

    # SUMij
    F[:p.nbeta,:,:] += np.einsum('i,jmn->imn',n[:p.nbeta],J[p.nbeta:p.nalpha,:,:]-0.5*K[p.nbeta:p.nalpha,:,:])
    F[p.nbeta:p.nalpha,:,:] += 0.5*np.einsum('jmn->mn',J[p.nbeta:p.nalpha,:,:]-K[p.nbeta:p.nalpha,:,:])
    F[p.nbeta:p.nalpha,:,:] -= 0.5*(J[p.nbeta:p.nalpha,:,:]-K[p.nbeta:p.nalpha,:,:]) #Remove diag.
    F[p.nalpha:p.nbf5,:,:] += np.einsum('i,jmn->imn',n[p.nalpha:p.nbf5],J[p.nbeta:p.nalpha,:,:]-0.5*K[p.nbeta:p.nalpha,:,:])
    
    # PRODWROij
    F[p.nbeta:p.nalpha,:,:] += np.einsum('j,jmn->mn',n[:p.nbeta],J[:p.nbeta,:,:]) - 0.5*np.einsum('j,jmn->mn',n[:p.nbeta],K[:p.nbeta,:,:])
    F[p.nbeta:p.nalpha,:,:] += np.einsum('j,jmn->mn',n[p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:]) - 0.5*np.einsum('j,jmn->mn',n[p.nalpha:p.nbf5],K[p.nalpha:p.nbf5,:,:])

    return F

def computeF_RO_GPU(J,K,n,H,cj12,ck12,p):

    # Matriz de Fock Generalizada
    F = cp.zeros((p.nbf5,p.nbf,p.nbf))

    ini = 0
    if(p.no1>1):
        ini = p.no1

    # nH
    F[:p.nbeta,:,:] += cp.einsum('i,mn->imn',n[:p.nbeta],H,optimize=True)                      # i = [1,nbf5]
    F[p.nbeta:p.nalpha,:,:] += 0.5*cp.array(H)                                                           # i = [nbeta,nalpha]
    F[p.nalpha:p.nbf5,:,:] += cp.einsum('i,mn->imn',n[p.nalpha:p.nbf5],H,optimize=True)        # i = [nalpha,nbf5]

    # nJ
    F[ini:p.nbeta,:,:] += cp.einsum('i,imn->imn',n[ini:p.nbeta],J[ini:p.nbeta,:,:],optimize=True)        # i = [ini,nbeta]
    F[p.nalpha:p.nbf5,:,:] += cp.einsum('i,imn->imn',n[p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:],optimize=True)  # i = [nalpha,nbf5]

    # C^J J
    np.fill_diagonal(cj12[ini:,ini:],0) # Remove diag.
    F[:p.nbeta,:,:] += cp.einsum('ij,jmn->imn',cj12[:p.nbeta,:p.nbeta],J[:p.nbeta,:,:],optimize=True)                               # i = [1,nbeta]
    F[:p.nbeta,:,:] += cp.einsum('ij,jmn->imn',cj12[:p.nbeta,p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:],optimize=True)                               # i = [1,nbeta]
    F[p.nalpha:p.nbf5,:,:] += cp.einsum('ij,jmn->imn',cj12[p.nalpha:p.nbf5,:p.nbeta],J[:p.nbeta,:,:],optimize=True)                                      # i = [nalpha,nbf5]
    F[p.nalpha:p.nbf5,:,:] += cp.einsum('ij,jmn->imn',cj12[p.nalpha:p.nbf5,p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:],optimize=True)                                      # i = [nalpha,nbf5]
    #F[ini:p.nbf5,:,:] -= np.einsum('ii,imn->imn',cj12[ini:p.nbf5,ini:p.nbf5],J[ini:p.nbf5,:,:],optimize=True) # quita i==j

    # -C^K K
    np.fill_diagonal(ck12[ini:,ini:],0) # Remove diag.
    F[:p.nbeta,:,:] -= cp.einsum('ij,jmn->imn',ck12[:p.nbeta,:p.nbeta],K[:p.nbeta,:,:],optimize=True)                                                # i = [1,nbeta]
    F[:p.nbeta,:,:] -= cp.einsum('ij,jmn->imn',ck12[:p.nbeta,p.nalpha:p.nbf5],K[p.nalpha:p.nbf5,:,:],optimize=True)                                                # i = [1,nbeta]
    F[p.nalpha:p.nbf5,:,:] -= cp.einsum('ij,jmn->imn',ck12[p.nalpha:p.nbf5,:p.nbeta],K[:p.nbeta,:,:],optimize=True)                                      # i = [nalpha,nbf5]
    F[p.nalpha:p.nbf5,:,:] -= cp.einsum('ij,jmn->imn',ck12[p.nalpha:p.nbf5,p.nalpha:p.nbf5],K[p.nalpha:p.nbf5,:,:],optimize=True)                                      # i = [nalpha,nbf5]
    #F[ini:p.nbf5,:,:] += np.einsum('ii,imn->imn',ck12[ini:p.nbf5,ini:p.nbf5],K[ini:p.nbf5,:,:],optimize=True) # quita i==j

    # SUMij
    F[:p.nbeta,:,:] += cp.einsum('i,jmn->imn',n[:p.nbeta],J[p.nbeta:p.nalpha,:,:]-0.5*K[p.nbeta:p.nalpha,:,:])
    F[p.nbeta:p.nalpha,:,:] += 0.5*cp.einsum('jmn->mn',J[p.nbeta:p.nalpha,:,:]-K[p.nbeta:p.nalpha,:,:])
    F[p.nbeta:p.nalpha,:,:] -= 0.5*cp.array((J[p.nbeta:p.nalpha,:,:]-K[p.nbeta:p.nalpha,:,:])) #Remove diag.
    F[p.nalpha:p.nbf5,:,:] += cp.einsum('i,jmn->imn',n[p.nalpha:p.nbf5],J[p.nbeta:p.nalpha,:,:]-0.5*K[p.nbeta:p.nalpha,:,:])

    # PRODWROij
    F[p.nbeta:p.nalpha,:,:] += cp.einsum('j,jmn->mn',n[:p.nbeta],J[:p.nbeta,:,:]) - 0.5*cp.einsum('j,jmn->mn',n[:p.nbeta],K[:p.nbeta,:,:])
    F[p.nbeta:p.nalpha,:,:] += cp.einsum('j,jmn->mn',n[p.nalpha:p.nbf5],J[p.nalpha:p.nbf5,:,:]) - 0.5*cp.einsum('j,jmn->mn',n[p.nalpha:p.nbf5],K[p.nalpha:p.nbf5,:,:])

    return F.get()

def computeLagrange(F,C,p):

    G = np.einsum('imn,ni->mi',F,C[:,0:p.nbf5],optimize=True)

    #Compute Lagrange multipliers
    elag = np.zeros((p.nbf,p.nbf))
    elag[0:p.noptorb,0:p.nbf5] = np.einsum('mi,mj->ij',C[:,0:p.noptorb],G,optimize=True)[0:p.noptorb,0:p.nbf5]

    return elag

def computeLagrange2(n,cj12,ck12,C,H,I,b_mnl,p):

    if p.RI:
        Hmat,b_MO = pynof.JKH_MO_tmp(C,H,I,b_mnl,p)
    else:
        Hmat,I_MO = pynof.JKH_MO_tmp(C,H,I,b_mnl,p)

    np.fill_diagonal(cj12,0) # Remove diag.
    np.fill_diagonal(ck12,0) # Remove diag.

    if p.gpu:
        elag = cp.zeros((p.nbf,p.nbf))
        n = cp.array(n)
        cj12 = cp.array(cj12)
        ck12 = cp.array(ck12)
        if p.RI:
            if(p.MSpin==0):
                # 2ndH/dy_ab
                elag[:,:p.nbf5] +=  cp.einsum('b,ab->ab',n,Hmat[:,:p.nbf5],optimize=True)

                # dJ_pp/dy_ab
                elag[:,:p.nbeta] +=  cp.einsum('b,abk,bbk->ab',n[:p.nbeta],b_MO[:,:p.nbeta,:],b_MO[:p.nbeta,:p.nbeta,:],optimize=True)
                elag[:,p.nalpha:p.nbf5] +=  cp.einsum('b,abk,bbk->ab',n[p.nalpha:p.nbf5],b_MO[:,p.nalpha:p.nbf5,:],b_MO[p.nalpha:p.nbf5,p.nalpha:p.nbf5,:],optimize=True)

                # C^J_pq dJ_pq/dy_ab 
                tmp = cp.einsum('bq,qqk->bk',cj12,b_MO[:p.nbf5,:p.nbf5,:],optimize=True)
                elag[:,:p.nbf5] +=  cp.einsum('abk,bk->ab',b_MO[:,:p.nbf5,:],tmp,optimize=True)

                # -C^K_pq dK_pq/dy_ab 
                elag[:,:p.nbf5] += -cp.einsum('bq,aqk,bqk->ab',ck12,b_MO[:,:p.nbf5,:],b_MO[:p.nbf5,:p.nbf5,:],optimize=True)
        else:
            if(p.MSpin==0):
                # 2ndH/dy_ab
                elag[:,:p.nbf5] +=  cp.einsum('b,ab->ab',n,Hmat[:,:p.nbf5],optimize=True)

                # dJ_pp/dy_ab
                elag[:,:p.nbeta] +=  cp.einsum('b,abbb->ab',n[:p.nbeta],I_MO[:,:p.nbeta,:p.nbeta,:p.nbeta],optimize=True)
                elag[:,p.nalpha:p.nbf5] +=  cp.einsum('b,abbb->ab',n[p.nalpha:p.nbf5],I_MO[:,p.nalpha:p.nbf5,p.nalpha:p.nbf5,p.nalpha:p.nbf5],optimize=True)

                # C^J_pq dJ_pq/dy_ab 
                elag[:,:p.nbf5] +=  cp.einsum('bq,abqq->ab',cj12,I_MO[:,:p.nbf5,:p.nbf5,:p.nbf5],optimize=True)

                # -C^K_pq dK_pq/dy_ab 
                elag[:,:p.nbf5] += -cp.einsum('bq,aqbq->ab',ck12,I_MO[:,:p.nbf5,:p.nbf5,:p.nbf5],optimize=True)
        return elag.get(),Hmat.get()
    else:
        elag = np.zeros((p.nbf,p.nbf))
        grad = np.zeros((p.nbf,p.nbf))
        if p.RI:
            if(p.MSpin==0):
                # 2ndH/dy_ab
                elag[:,:p.nbf5] +=  np.einsum('b,ab->ab',n,Hmat[:,:p.nbf5],optimize=True)

                # dJ_pp/dy_ab
                elag[:,:p.nbeta] +=  np.einsum('b,abk,bbk->ab',n[:p.nbeta],b_MO[:,:p.nbeta,:],b_MO[:p.nbeta,:p.nbeta,:],optimize=True)
                elag[:,p.nalpha:p.nbf5] +=  np.einsum('b,abk,bbk->ab',n[p.nalpha:p.nbf5],b_MO[:,p.nalpha:p.nbf5,:],b_MO[p.nalpha:p.nbf5,p.nalpha:p.nbf5,:],optimize=True)

                # C^J_pq dJ_pq/dy_ab
                tmp = np.einsum('bq,qqk->bk',cj12,b_MO[:p.nbf5,:p.nbf5,:],optimize=True)
                elag[:,:p.nbf5] +=  np.einsum('abk,bk->ab',b_MO[:,:p.nbf5,:],tmp,optimize=True)

                # -C^K_pq dK_pq/dy_ab 
                elag[:,:p.nbf5] += -np.einsum('bq,aqk,bqk->ab',ck12,b_MO[:,:p.nbf5,:],b_MO[:p.nbf5,:p.nbf5,:],optimize=True)
        else:
            if(p.MSpin==0):
                # 2ndH/dy_ab
                elag[:,:p.nbf5] +=  np.einsum('b,ab->ab',n,Hmat[:,:p.nbf5],optimize=True)

                # dJ_pp/dy_ab
                elag[:,:p.nbeta] +=  np.einsum('b,abbb->ab',n[:p.nbeta],I_MO[:,:p.nbeta,:p.nbeta,:p.nbeta],optimize=True)
                elag[:,p.nalpha:p.nbf5] +=  np.einsum('b,abbb->ab',n[p.nalpha:p.nbf5],I_MO[:,p.nalpha:p.nbf5,p.nalpha:p.nbf5,p.nalpha:p.nbf5],optimize=True)

                # C^J_pq dJ_pq/dy_ab 
                elag[:,:p.nbf5] +=  np.einsum('bq,abqq->ab',cj12,I_MO[:,:p.nbf5,:p.nbf5,:p.nbf5],optimize=True)

                # -C^K_pq dK_pq/dy_ab 
                elag[:,:p.nbf5] += -np.einsum('bq,aqbq->ab',ck12,I_MO[:,:p.nbf5,:p.nbf5,:p.nbf5],optimize=True)

    return elag,Hmat

def computeE_elec(H,C,n,elag,p):
    #EELECTRr
    E = 0

    E = E + np.einsum('ii',elag[:p.nbf5,:p.nbf5],optimize=True)
    E = E + np.einsum('i,mi,mn,ni',n[:p.nbeta],C[:,:p.nbeta],H,C[:,:p.nbeta],optimize=True)
    if(not p.HighSpin):
        E = E + np.einsum('i,mi,mn,ni',n[p.nbeta:p.nalpha],C[:,p.nbeta:p.nalpha],H,C[:,p.nbeta:p.nalpha],optimize=True)
    elif(p.HighSpin):
        E = E + 0.5*np.einsum('mi,mn,ni',C[:,p.nbeta:p.nalpha],H,C[:,p.nbeta:p.nalpha],optimize=True)

    E = E + np.einsum('i,mi,mn,ni',n[p.nalpha:p.nbf5],C[:,p.nalpha:p.nbf5],H,C[:,p.nalpha:p.nbf5],optimize=True)

    return E


def computeLagrangeConvergency(elag):
    # Convergency

    sumdiff = np.sum(np.abs(elag-elag.T))
    maxdiff = np.max(np.abs(elag-elag.T))

    return sumdiff,maxdiff


def ENERGY1r(C,n,H,I,b_mnl,cj12,ck12,p):

    if(p.no1==0):
        elag,Hmat = pynof.computeLagrange2(n,cj12,ck12,C,H,I,b_mnl,p)
        E = np.einsum('ii',elag[:p.nbf5,:p.nbf5],optimize=True)
        E = E + np.einsum('i,ii',n[:p.nbf5],Hmat[:p.nbf5,:p.nbf5],optimize=True)
    else:
        J,K = pynof.computeJKj(C,I,b_mnl,p)
        if(p.MSpin==0):
            F = computeF_RC_driver(J,K,n,H,cj12,ck12,p)
        elif(not p.MSpin==0):
            F = computeF_RO_driver(J,K,n,H,cj12,ck12,p)
        elag = computeLagrange(F,C,p)
        E = computeE_elec(H,C,n,elag,p)

    sumdiff,maxdiff = computeLagrangeConvergency(elag)

    return E,elag,sumdiff,maxdiff

def fmiug_scaling(fmiug0,elag,i_ext,nzeros,nbf,noptorb):

    #scaling
    fmiug = np.zeros((nbf,nbf))
    if(i_ext == 0 and fmiug0 is None):
        fmiug[:noptorb,:noptorb] = ((elag[:noptorb,:noptorb] + elag[:noptorb,:noptorb].T) / 2)

    else:
        fmiug[:noptorb,:noptorb] = (elag[:noptorb,:noptorb] - elag[:noptorb,:noptorb].T)
        fmiug = np.tril(fmiug,-1) + np.tril(fmiug,-1).T
        for k in range(nzeros+9+1):
            fmiug[(abs(fmiug) > 10**(9-k)) & (abs(fmiug) < 10**(10-k))] *= 0.1
        np.fill_diagonal(fmiug[:noptorb,:noptorb],fmiug0[:noptorb])

    return fmiug

@njit
def fmiug_diis(fk,fmiug,idiis,bdiis,cdiis,maxdiff,noptorb,ndiis,perdiis):

    fk[idiis,0:noptorb,0:noptorb] = fmiug[0:noptorb,0:noptorb]
    for m in range(idiis+1):
        bdiis[m][idiis] = 0
        for i in range(noptorb):
            for j in range(i):
                bdiis[m][idiis] = bdiis[m][idiis] + fk[m][i][j]*fk[idiis][j][i]
        bdiis[idiis][m] = bdiis[m][idiis]
        bdiis[m][idiis+1] = -1.
        bdiis[idiis+1][m] = -1.
    bdiis[idiis+1][idiis+1] = 0.

    if(idiis>=ndiis):
        cdiis = np.zeros((idiis+2))
        cdiis[0:idiis+1] = 0.
        cdiis[idiis+1] = -1.
        x = np.linalg.solve(bdiis[0:idiis+2,0:idiis+2],cdiis[0:idiis+2])

        for i in range(noptorb):
            for j in range(i):
                fmiug[i][j] = 0.
                for k in range(idiis+1):
                    fmiug[i][j] = fmiug[i][j] + x[k]*fk[k][i][j]
                fmiug[j][i] = fmiug[i][j]
#        print("idiis", idiis)
    if(idiis>=ndiis and perdiis):
        idiis = 0
    else:
        idiis = idiis + 1

    return fk,fmiug,idiis,bdiis

def check_ortho(C,S,p):

    # Revisa ortonormalidad
    orthonormality = True
    CTSC = np.matmul(np.matmul(np.transpose(C),S),C)
    ortho_deviation = np.abs(CTSC - np.identity(p.nbf))
    if (np.any(ortho_deviation > 10**-6)):
        orthonormality = False
    if not orthonormality:
        print("Orthonormality violations {:d}, Maximum Violation {:f}".format((ortho_deviation > 10**-6).sum(),ortho_deviation.max()))
        print("Trying to orthonormalize")
        C = orthonormalize(C,S)
        C = check_ortho(C,S,p)
    else:
        print("No violations of the orthonormality")
    for j in range(p.nbf):
        #Obtiene el índice del coeficiente con mayor valor absoluto del MO
        idxmaxabsval = 0
        for i in range(p.nbf):
            if(abs(C[i][j])>abs(C[idxmaxabsval][j])):
                 idxmaxabsval = i
    # Ajusta el signo del MO
    sign = np.sign(C[idxmaxabsval][j])
    C[0:p.nbf,j] = sign*C[0:p.nbf,j]

    return C

def orthonormalize(C,S):
    eigval,eigvec = eigh(S) 
    S_12 = np.einsum('ij,j->ij',eigvec,eigval**(-1/2),optimize=True)

    Cnew = np.einsum('ik,kj->ij',S,C,optimize=True)

    Cnew2 = np.einsum('ki,kj->ij',S_12,Cnew)

    for i in range(Cnew2.shape[1]):
        norm = np.einsum('k,k->',Cnew2[:,i],Cnew2[:,i],optimize=True)
        Cnew2[:,i] = Cnew2[:,i]/np.sqrt(norm)
        for j in range(i+1,Cnew2.shape[1]):
            val = -np.einsum('k,k->',Cnew2[:,i],Cnew2[:,j],optimize=True)
            Cnew2[:,j] = Cnew2[:,j] + val*Cnew2[:,i]

    C = np.einsum("ik,kj->ij",S_12,Cnew2,optimize=True)

    return C

def rotate_orbital(y,C,p):

    ynew = np.zeros((p.nbf,p.nbf))

    n = 0
    for i in range(p.nbf5):
        for j in range(i+1,p.nbf):
            ynew[i,j] =  y[n]
            ynew[j,i] = -y[n]
            n += 1

    U = expm(ynew)

    Cnew = np.einsum("mr,rp->mp",C,U,optimize=True)

    return Cnew


@njit(cache=True)
def extract_tiu_tensor(t,k):
    dim = len(t)
    var = int(dim*(dim-1)/2)
    t_extracted = np.zeros((var,var))
    i = 0
    for p in range(dim):
        for q in range(p+k,dim):
            j = 0
            for r in range(dim):
                for s in range(r+k,dim):
                    t_extracted[i,j] = t[p,q,r,s]
                    j += 1
            i += 1

    return t_extracted

def perturb_solution(C,gamma,grad_orb,grad_occ,p):
    grad_orb = pynof.perturb_gradient(grad_orb,p.tol_gorb)
    y = -grad_orb
    C = pynof.rotate_orbital(y,C,p)
    grad_occ = pynof.perturb_gradient(grad_occ,p.tol_gocc)
    gamma += -grad_occ

    return C,gamma

@njit
def perturb_gradient(grad,tol):
    dim = grad.shape[0]
    for i in range(dim):
        if(np.abs(grad[i])<tol):
            grad[i] = np.sign(grad[i])*tol

    return grad

@njit(cache=True)
def n_to_gammas_trigonometric(n,nv,no1,ndoc,ndns,ncwo):
    gamma = np.zeros(nv)
    for i in range(ndoc):
        idx = no1 + i
        gamma[i] = np.arccos(np.sqrt(2.0*n[idx]-1.0))
        prefactor = max(1-n[idx],1e-14)
        for j in range(ncwo-1):
            jg = ndoc + i*(ncwo-1) + j
            ig = no1 + ndns + ncwo*(ndoc - i - 1) + j
            gamma[jg] = np.arcsin(np.sqrt(n[ig]/prefactor))
            prefactor = prefactor * (np.cos(gamma[jg]))**2
    return gamma

@njit(cache=True)
def n_to_gammas_softmax(n,nv,no1,ndoc,ndns,ncwo):

    gamma = np.zeros((nv))

    for i in range(ndoc):

        ll = no1 + ndns + ncwo*(ndoc - i - 1)
        ul = no1 + ndns + ncwo*(ndoc - i)

        llg = ll - ndns + ndoc - no1
        ulg = ul - ndns + ndoc - no1

        ns = n[ll:ul]

        A = np.zeros((ncwo,ncwo))
        b = np.zeros((ncwo))

        for j in range(ncwo):
            A[j,:] = ns[j]
            A[j,j] = ns[j]-1
            b[j] = -ns[j]

        x = np.log(np.linalg.solve(A,b))
        gamma[llg:ulg] = x

    return gamma

@njit(cache=True)
def compute_gammas_trigonometric(nv,ndoc,ncwo):
    gamma = np.zeros((nv))
    for i in range(ndoc):
        gamma[i] = np.arccos(np.sqrt(2.0*0.999-1.0))
        for j in range(ncwo-1):
            ig = ndoc+i*(ncwo-1)+j
            gamma[ig] = np.arcsin(np.sqrt(1.0/(ncwo-j)))
    return gamma



import pynof
import numpy as np
from numba import prange,njit
try:
    import cupy as cp
except:
    pass

#CJCKD5
@njit(parallel=True, cache=True)
def CJCKD5(n,no1,ndoc,nsoc,nbeta,nalpha,ndns,ncwo,MSpin):

    # Interpair Electron correlation #
    cj12 = 2*np.outer(n,n)
    ck12 = np.outer(n,n)

    # Intrapair Electron Correlation

    if(MSpin==0 and nsoc>1):
        ck12[nbeta:nalpha,nbeta:nalpha] = 2*np.outer(n[nbeta:nalpha],n[nbeta:nalpha])

    for l in prange(ndoc):
        ldx = no1 + l
        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = no1 + ndns + ncwo*(ndoc - l - 1)
        ul = no1 + ndns + ncwo*(ndoc - l)

        cj12[ldx,ll:ul] = 0
        cj12[ll:ul,ldx] = 0

        cj12[ll:ul,ll:ul] = 0

        ck12[ldx,ll:ul] = np.sqrt(n[ldx]*n[ll:ul])
        ck12[ll:ul,ldx] = np.sqrt(n[ldx]*n[ll:ul])

        ck12[ll:ul,ll:ul] = -np.sqrt(np.outer(n[ll:ul],n[ll:ul]))

    return cj12,ck12

@njit(parallel=True, cache=True)
def der_CJCKD5(n,ista,dn_dgamma,no1,ndoc,nalpha,nbeta,nv,nbf5,ndns,ncwo):

    # Interpair Electron correlation #

    Dcj12r = np.zeros((nbf5,nbf5,nv))
    Dck12r = np.zeros((nbf5,nbf5,nv))
    for k in prange(nv):
        Dcj12r[:,:,k] = 2*np.outer(dn_dgamma[:,k],n)
        Dck12r[:,:,k] = np.outer(dn_dgamma[:,k],n)
    #Dcj12r = 2*np.einsum('ik,j->ijk',dn_dgamma,n)    
    #Dck12r = np.einsum('ik,j->ijk',dn_dgamma,n)

    # Intrapair Electron Correlation

    for l in prange(ndoc):
        ldx = no1 + l

        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = no1 + ndns + ncwo*(ndoc - l - 1)
        ul = no1 + ndns + ncwo*(ndoc - l)

        Dcj12r[ldx,ll:ul,:nv] = 0
        Dcj12r[ll:ul,ldx,:nv] = 0

        Dcj12r[ll:ul,ll:ul,:nv] = 0

        a = max(n[ldx],10**-15)
        b = n[ll:ul]
        b[b<10**-15] = 10**-15

        for k in range(nv):
            Dck12r[ldx,ll:ul,k] = 1/2 * 1/np.sqrt(a) * dn_dgamma[ldx,k] * np.sqrt(n[ll:ul])
            Dck12r[ll:ul,ldx,k] = 1/2 * 1/np.sqrt(b) * dn_dgamma[ll:ul,k] * np.sqrt(n[ldx])
            Dck12r[ll:ul,ll:ul,k] = - 1/2 * np.outer(1/np.sqrt(b)*dn_dgamma[ll:ul,k],np.sqrt(n[ll:ul]))
        #Dck12r[ldx,ll:ul,:nv] = 1/2 * 1/np.sqrt(a) * np.einsum('j,i->ij',dn_dgamma[ldx,:nv],np.sqrt(n[ll:ul]))
        #Dck12r[ll:ul,ldx,:nv] = 1/2 * np.einsum('i,ij->ij', 1/np.sqrt(b),dn_dgamma[ll:ul,:nv])*np.sqrt(n[ldx])

        #for k in range(nv):
        #    Dck12r[ll:ul,ll:ul,k] = - 1/2 * np.einsum('i,i,j->ij',1/np.sqrt(b),dn_dgamma[ll:ul,k],np.sqrt(n[ll:ul]))

    return Dcj12r,Dck12r

#CJCKD7
@njit(parallel=True, cache=True)
def CJCKD7(n,ista,no1,ndoc,nsoc,nbeta,nalpha,ndns,ncwo,MSpin):

    if(ista==0):
        fi = n*(1-n)
        fi[fi<=0] = 0
        fi = np.sqrt(fi)      
    else:
        fi = 2*n*(1-n)

    # Interpair Electron correlation #

    #cj12 = 2*np.einsum('i,j->ij',n,n)
    #ck12 = np.einsum('i,j->ij',n,n) + np.einsum('i,j->ij',fi,fi)
    cj12 = 2*np.outer(n,n)
    ck12 = np.outer(n,n) + np.outer(fi,fi)
    
    # Intrapair Electron Correlation

    if(MSpin==0 and nsoc>1):
        ck12[nbeta:nalpha,nbeta:nalpha] = 2*np.outer(n[nbeta:nalpha],n[nbeta:nalpha])

    for l in prange(ndoc):            
        ldx = no1 + l
        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = no1 + ndns + ncwo*(ndoc - l - 1)
        ul = no1 + ndns + ncwo*(ndoc - l)

        cj12[ldx,ll:ul] = 0    
        cj12[ll:ul,ldx] = 0    
    
        cj12[ll:ul,ll:ul] = 0    
        
        ck12[ldx,ll:ul] = np.sqrt(n[ldx]*n[ll:ul])
        ck12[ll:ul,ldx] = np.sqrt(n[ldx]*n[ll:ul])

        ck12[ll:ul,ll:ul] = -np.sqrt(np.outer(n[ll:ul],n[ll:ul]))

    return cj12,ck12        
       
@njit(parallel=True, cache=True)
def der_CJCKD7(n,ista,dn_dgamma,no1,ndoc,nalpha,nbeta,nv,nbf5,ndns,ncwo):

    if(ista==0):
        fi = n*(1-n)
        fi[fi<=0] = 0
        fi = np.sqrt(fi)      
    else:
        fi = 2*n*(1-n)
            
    dfi_dgamma = np.zeros((nbf5,nv))
    for i in prange(no1,nbf5):
        a = max(fi[i],10**-15)
        for k in range(nv):
            if(ista==0):
                dfi_dgamma[i,k] = 1/(2*a)*(1-2*n[i])*dn_dgamma[i][k]
            else:
                dfi_dgamma[i,k] = 2*(1-2*n[i])*dn_dgamma[i][k]
                
   
    # Interpair Electron correlation #

    Dcj12r = np.zeros((nbf5,nbf5,nv))
    Dck12r = np.zeros((nbf5,nbf5,nv))
    for k in prange(nv):
        Dcj12r[:,:,k] = 2*np.outer(dn_dgamma[:,k],n)    
        Dck12r[:,:,k] = np.outer(dn_dgamma[:,k],n) + np.outer(dfi_dgamma[:,k],fi)
    #Dcj12r = 2*np.einsum('ik,j->ijk',dn_dgamma,n)    
    #Dck12r = np.einsum('ik,j->ijk',dn_dgamma,n) + np.einsum('ik,j->ijk',dfi_dgamma,fi)    

    # Intrapair Electron Correlation

    for l in prange(ndoc):            
        ldx = no1 + l

        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = no1 + ndns + ncwo*(ndoc - l - 1)
        ul = no1 + ndns + ncwo*(ndoc - l)

        Dcj12r[ldx,ll:ul,:nv] = 0
        Dcj12r[ll:ul,ldx,:nv] = 0

        Dcj12r[ll:ul,ll:ul,:nv] = 0   

        a = max(n[ldx],10**-15)
        b = n[ll:ul]
        b[b<10**-15] = 10**-15        
        
        for k in range(nv):
            Dck12r[ldx,ll:ul,k] = 1/2 * 1/np.sqrt(a) * dn_dgamma[ldx,k] * np.sqrt(n[ll:ul])
            Dck12r[ll:ul,ldx,k] = 1/2 * 1/np.sqrt(b) * dn_dgamma[ll:ul,k] * np.sqrt(n[ldx])
            Dck12r[ll:ul,ll:ul,k] = - 1/2 * np.outer(1/np.sqrt(b)*dn_dgamma[ll:ul,k],np.sqrt(n[ll:ul]))
        #Dck12r[ldx,ll:ul,:nv] = 1/2 * 1/np.sqrt(a) * np.einsum('j,i->ij',dn_dgamma[ldx,:nv],np.sqrt(n[ll:ul]))
        #Dck12r[ll:ul,ldx,:nv] = 1/2 * np.einsum('i,ij->ij', 1/np.sqrt(b),dn_dgamma[ll:ul,:nv])*np.sqrt(n[ldx])
        
        #for k in range(nv):
        #    Dck12r[ll:ul,ll:ul,k] = - 1/2 * np.einsum('i,i,j->ij',1/np.sqrt(b),dn_dgamma[ll:ul,k],np.sqrt(n[ll:ul]))
                        
    return Dcj12r,Dck12r

#CJCKD8
#@njit(parallel=True, cache=True)
#def CJCKD8(n,ista,no1,ndoc,nsoc,nbeta,nalpha,ndns,ncwo,MSpin,lamb):
#
#    nbf5 = len(n)
#    delta = np.zeros((nbf5,nbf5))
#    for i in range(nbf5):
#        for j in range(nbf5):
#            delta[i,j] = lamb*min(n[i]*n[j],(1-n[i])*(1-n[j]))
#    pi = np.zeros((nbf5,nbf5))
#    for i in range(nbf5):
#        for j in range(nbf5):
#            pi[i,j] = np.sqrt((n[i]*(1-n[j])+delta[i,j]) * (n[j]*(1-n[i])+delta[j,i]))
#
#    # Interpair Electron correlation #
#
#    #cj12 = 2*np.einsum('i,j->ij',n,n)
#    #ck12 = np.einsum('i,j->ij',n,n) + np.einsum('i,j->ij',fi,fi)
#    cj12 = 2*np.outer(n,n) - delta
#    ck12 = np.outer(n,n) - delta + pi
#
#    # Intrapair Electron Correlation
#
#    if(MSpin==0 and nsoc>1):
#        ck12[nbeta:nalpha,nbeta:nalpha] = 2*np.outer(n[nbeta:nalpha],n[nbeta:nalpha])
#
#    for l in range(ndoc):
#        ldx = no1 + l
#        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
#        ll = no1 + ndns + ncwo*(ndoc - l - 1)
#        ul = no1 + ndns + ncwo*(ndoc - l)
#
#        cj12[ldx,ll:ul] = 0
#        cj12[ll:ul,ldx] = 0
#
#        cj12[ll:ul,ll:ul] = 0
#
#        ck12[ldx,ll:ul] = np.sqrt(n[ldx]*n[ll:ul])
#        ck12[ll:ul,ldx] = np.sqrt(n[ldx]*n[ll:ul])
#
#        ck12[ll:ul,ll:ul] = -np.sqrt(np.outer(n[ll:ul],n[ll:ul]))
#
#    return cj12,ck12


#CJCKD8
@njit(parallel=True, cache=True)
def CJCKD8(n,no1,ndoc,nsoc,nbeta,nalpha,ndns,ncwo,MSpin):

    h_cut = 0.02*np.sqrt(2.0)
    n_d = np.zeros((len(n)))

    for i in prange(ndoc):
        idx = no1 + i
        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = no1 + ndns + ncwo*(ndoc - i - 1)
        ul = no1 + ndns + ncwo*(ndoc - i)
        h = 1.0 - n[idx]
        coc = h / h_cut
        arg = -coc ** 2
        F = np.exp(arg)  # ! Hd/Hole
        n_d[idx] = n[idx] * F
        n_d[ll:ul] = n[ll:ul] * F  # ROd = RO*Hd/Hole

    n_d12 = np.sqrt(n_d)
    fi = n*(1-n)
    fi[fi<=0] = 0
    fi = np.sqrt(fi)

    # Interpair Electron correlation #

    cj12 = 2*np.outer(n,n)
    ck12 = np.outer(n,n)

    ck12[no1:nbeta,nalpha:] += np.outer(fi[no1:nbeta],fi[nalpha:])
    ck12[nalpha:,no1:nbeta] += np.outer(fi[nalpha:],fi[no1:nbeta])
    ck12[nalpha:,nalpha:] += np.outer(fi[nalpha:],fi[nalpha:])

    # Intrapair Electron Correlation

    if(MSpin==0 and nsoc>0):
        half = np.zeros((nsoc)) + 0.5
        ck12[no1:nbeta,nbeta:nalpha] += 0.5*np.outer(fi[no1:nbeta],half)
        ck12[nbeta:nalpha,no1:nbeta] += 0.5*np.outer(half,fi[no1:nbeta])
        ck12[nbeta:nalpha,nalpha:] += np.outer(half,fi[nalpha:])
        ck12[nalpha:,nbeta:nalpha] += np.outer(fi[nalpha:],half)

    if(MSpin==0 and nsoc>1): #then
        ck12[nbeta:nalpha,nbeta:nalpha] = 0.5

    ck12[no1:nbeta,nalpha:] += np.outer(n_d12[no1:nbeta],n_d12[nalpha:]) - np.outer(n_d[no1:nbeta],n_d[nalpha:])
    ck12[nalpha:,no1:nbeta] += np.outer(n_d12[nalpha:],n_d12[no1:nbeta]) - np.outer(n_d[nalpha:],n_d[no1:nbeta])
    ck12[nalpha:,nalpha:] += -np.outer(n_d12[nalpha:],n_d12[nalpha:]) - np.outer(n_d[nalpha:],n_d[nalpha:])

    for l in prange(ndoc):
        ldx = no1 + l
        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = no1 + ndns + ncwo*(ndoc - l - 1)
        ul = no1 + ndns + ncwo*(ndoc - l)

        cj12[ldx,ll:ul] = 0
        cj12[ll:ul,ldx] = 0

        cj12[ll:ul,ll:ul] = 0

        ck12[ldx,ll:ul] = np.sqrt(n[ldx]*n[ll:ul])
        ck12[ll:ul,ldx] = np.sqrt(n[ldx]*n[ll:ul])

        ck12[ll:ul,ll:ul] = -np.sqrt(np.outer(n[ll:ul],n[ll:ul]))

    return cj12,ck12

@njit(parallel=True, cache=True)
def der_CJCKD8(n,dn_dgamma,no1,ndoc,nalpha,nbeta,nv,nbf5,ndns,ncwo,MSpin,nsoc):

    h_cut = 0.02*np.sqrt(2.0)
    n_d = np.zeros((len(n)))
    dn_d_dgamma = np.zeros((len(n),nv))
    dn_d12_dgamma = np.zeros((len(n),nv))

    for i in prange(ndoc):
        idx = no1 + i
        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = no1 + ndns + ncwo*(ndoc - i - 1)
        ul = no1 + ndns + ncwo*(ndoc - i)
        h_idx = 1.0-n[idx]
        coc = h_idx/h_cut
        arg = -coc**2
        F_idx = np.exp(arg)                # Hd/Hole
        n_d[idx] = n[idx] * F_idx
        n_d[ll:ul] = n[ll:ul] * F_idx      # n_d = RO*Hd/Hole
        dn_d_dgamma[idx,:] = F_idx*dn_dgamma[idx,:] * (1-n[idx]*(- 2*coc/h_cut))
        dn_d_dgamma[ll:ul,:] = F_idx*(dn_dgamma[ll:ul,:] - (- 2*coc/h_cut)*np.outer(n[ll:ul],dn_dgamma[idx,:]))

    n_d12 = np.sqrt(n_d)
    for i in prange(nbf5):
        dn_d12_dgamma[i, :] = 0.5 * dn_d_dgamma[i, :] / max(n_d12[i], 1e-15)

    fi = n*(1-n)
    fi[fi<=0] = 0
    fi = np.sqrt(fi)


    dfi_dgamma = np.zeros((nbf5,nv))
    for i in prange(no1,nbf5):
        a = max(fi[i],10**-15)
        for k in range(nv):
            dfi_dgamma[i,k] = 1/(2*a)*(1-2*n[i])*dn_dgamma[i][k]

    # Interpair Electron correlation #

    Dcj12r = np.zeros((nbf5,nbf5,nv))
    Dck12r = np.zeros((nbf5,nbf5,nv))
    for k in prange(nv):
        Dcj12r[:,:,k] = 2*np.outer(dn_dgamma[:,k],n)
        Dck12r[:,:,k] = np.outer(dn_dgamma[:,k],n)

    #Dcj12r = 2*np.einsum('ik,j->ijk',dn_dgamma,n)
    #Dck12r = np.einsum('ik,j->ijk',dn_dgamma,n)

    for i in prange(nalpha,nbf5):
        Dck12r[no1:nbeta,i,:] += dfi_dgamma[no1:nbeta,:]*fi[i]
        Dck12r[i,no1:nbeta,:] += np.outer(fi[no1:nbeta],dfi_dgamma[i,:])
        Dck12r[i,nalpha:nbf5,:] += np.outer(fi[nalpha:nbf5],dfi_dgamma[i,:])

    if(MSpin==0 and nsoc>0):
        for i in prange(nbeta,nalpha):
            Dck12r[no1:nbeta,i,:] += 0.5*dfi_dgamma[no1:nbeta,:]*0.5
            Dck12r[nalpha:nbf5,i,:] += dfi_dgamma[nalpha:nbf5,:]*0.5

    if(MSpin==0 and nsoc>1):
        Dck12r[nbeta:nalpha, nbeta:nalpha, :] = 0.0

    for i in prange(nalpha,nbf5):
        Dck12r[no1:nbeta,i,:] += dn_d12_dgamma[no1:nbeta,:]*n_d12[i] - dn_d_dgamma[no1:nbeta,:]*n_d[i]
        Dck12r[i,no1:nbeta,:] += np.outer(n_d12[no1:nbeta],dn_d12_dgamma[i,:]) - np.outer(n_d[no1:nbeta],dn_d_dgamma[i,:])
        Dck12r[i,nalpha:nbf5,:] += -np.outer(n_d12[nalpha:nbf5],dn_d12_dgamma[i,:]) - np.outer(n_d[nalpha:nbf5],dn_d_dgamma[i,:])

    # Intrapair Electron Correlation

    for l in prange(ndoc):
        ldx = no1 + l

        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = no1 + ndns + ncwo*(ndoc - l - 1)
        ul = no1 + ndns + ncwo*(ndoc - l)

        Dcj12r[ldx,ll:ul,:nv] = 0
        Dcj12r[ll:ul,ldx,:nv] = 0

        Dcj12r[ll:ul,ll:ul,:nv] = 0

        a = max(n[ldx],10**-15)
        b = n[ll:ul]
        b[b<10**-15] = 10**-15

        for k in range(nv):
            Dck12r[ldx,ll:ul,k] = 1/2 * 1/np.sqrt(a) * dn_dgamma[ldx,k] * np.sqrt(n[ll:ul])
            Dck12r[ll:ul,ldx,k] = 1/2 * 1/np.sqrt(b) * dn_dgamma[ll:ul,k] * np.sqrt(n[ldx])
            Dck12r[ll:ul,ll:ul,k] = - 1/2 * np.outer(1/np.sqrt(b)*dn_dgamma[ll:ul,k],np.sqrt(n[ll:ul]))

    return Dcj12r,Dck12r

def compute_2RDM(pp,n,orb_pair_ll,orb_pair_ul):

    norb = len(n)
    # PNOF5
    # Interpair Electron correlation #
    Id = np.identity(norb)

    inter = np.outer(n,n)
    intra = 0*np.outer(n,n)

    # Intrapair Electron Correlation
    for l in prange(pp.ndoc):
        ldx = pp.no1 + l
        # inicio y fin de los orbitales acoplados a los fuertemente ocupados
        ll = orb_pair_ll[l]#pp.no1 + pp.ndns + pp.ncwo*(pp.ndoc - l - 1)
        ul = orb_pair_ul[l]#pp.no1 + pp.ndns + pp.ncwo*(pp.ndoc - l)

        inter[ldx,ldx] = 0
        inter[ldx,ll:ul] = 0
        inter[ll:ul,ldx] = 0
        inter[ll:ul,ll:ul] = 0

        intra[ldx,ldx] = np.sqrt(n[ldx]*n[ldx])
        intra[ldx,ll:ul] = -np.sqrt(n[ldx]*n[ll:ul])
        intra[ll:ul,ldx] = -np.sqrt(n[ldx]*n[ll:ul])
        intra[ll:ul,ll:ul] = np.sqrt(np.outer(n[ll:ul],n[ll:ul]))

    for i in range(pp.nbeta,pp.nalpha):
        inter[i,i] = 0

    Daa = np.einsum('pq,pr,qt->pqrt',inter,Id,Id,optimize=True) - np.einsum('pq,pt,qr->pqrt',inter,Id,Id,optimize=True)
    Dab = np.einsum('pq,pr,qt->pqrt',inter,Id,Id,optimize=True) + np.einsum('pr,pq,rt->pqrt',intra,Id,Id,optimize=True)

    # PNOF7
    if(pp.ipnof == 7 or pp.ipnof==8):
        fi = n*(1-n)
        fi[fi<=0] = 0
        fi = np.sqrt(fi)
        Pi_s = np.outer(fi,fi)
        # Intrapair Electron Correlation
        for l in prange(pp.ndoc):
            ldx = pp.no1 + l
            # inicio y fin de los orbitales acoplados a los fuertemente ocupados
            ll = orb_pair_ll[l]#pp.no1 + pp.ndns + pp.ncwo*(pp.ndoc - l - 1)
            ul = orb_pair_ul[l]#pp.no1 + pp.ndns + pp.ncwo*(pp.ndoc - l)

            Pi_s[ldx,ldx] = 0
            Pi_s[ldx,ll:ul] = 0
            Pi_s[ll:ul,ldx] = 0
            Pi_s[ll:ul,ll:ul] = 0
        for i in range(pp.nbeta,pp.nalpha):
            Pi_s[i,i] = 0

        inter2 = 0*Pi_s.copy()
        inter2[pp.nbeta:pp.nalpha,pp.nbeta:pp.nalpha] = Pi_s[pp.nbeta:pp.nalpha,pp.nbeta:pp.nalpha]
        Pi_s[pp.nbeta:pp.nalpha,pp.nbeta:pp.nalpha] = 0

        Dab -= np.einsum('pr,pt,qr->pqrt',inter2,Id,Id,optimize=True)
#
#
#    #inter3 = 0*inter.copy()
#    #inter3[pp.nbeta:pp.nalpha,pp.nbeta:pp.nalpha] = -1/4
#    # inter3 = inter2?

        if(pp.ipnof==8):
            Pi_s[:pp.nbeta,:pp.nbeta] = 0
            Pi_s[:pp.nbeta,pp.nbeta:pp.nalpha] *= 0.5
            Pi_s[pp.nbeta:pp.nalpha,:pp.nbeta] *= 0.5

        Dab -= np.einsum('pr,pq,rt->pqrt',Pi_s,Id,Id,optimize=True)

        if(pp.ipnof==8):

            h_cut = 0.02*np.sqrt(2.0)
            n_d = np.zeros((len(n)))

            for i in prange(pp.ndoc):
                idx = pp.no1 + i
                # inicio y fin de los orbitales acoplados a los fuertemente ocupados
                ll = orb_pair_ll[i]#pp.no1 + pp.ndns + pp.ncwo*(pp.ndoc - i - 1)
                ul = orb_pair_ul[i]#pp.no1 + pp.ndns + pp.ncwo*(pp.ndoc - i)
                h = 1.0 - n[idx]
                coc = h / h_cut
                arg = -coc ** 2
                F = np.exp(arg)  # ! Hd/Hole
                n_d[idx] = n[idx] * F
                n_d[ll:ul] = n[ll:ul] * F  # ROd = RO*Hd/Hole

            n_d12 = np.sqrt(n_d)

            inter = np.outer(n_d12,n_d12) - np.outer(n_d,n_d)
            inter2 = np.outer(n_d12,n_d12) + np.outer(n_d,n_d)
            # Intrapair Electron Correlation
            for l in prange(pp.ndoc):
                ldx = pp.no1 + l
                # inicio y fin de los orbitales acoplados a los fuertemente ocupados
                ll = orb_pair_ll[l]#pp.no1 + pp.ndns + pp.ncwo*(pp.ndoc - l - 1)
                ul = orb_pair_ul[l]#pp.no1 + pp.ndns + pp.ncwo*(pp.ndoc - l)

                inter[ldx,ldx] = 0
                inter[ldx,ll:ul] = 0
                inter[ll:ul,ldx] = 0
                inter[ll:ul,ll:ul] = 0
                inter2[ldx,ldx] = 0
                inter2[ldx,ll:ul] = 0
                inter2[ll:ul,ldx] = 0
                inter2[ll:ul,ll:ul] = 0

            inter[pp.nbeta:,pp.nbeta:] = 0
            inter[:pp.nalpha,:pp.nalpha] = 0
            inter2[pp.nbeta:,pp.nbeta:] = 0
            inter2[:pp.nalpha,:] = 0
            inter2[:,:pp.nalpha] = 0

            Pi_d = inter - inter2

            Dab -= np.einsum('pr,pq,rt->pqrt',Pi_d,Id,Id,optimize=True)

    return Daa, Dab

# Creamos un seleccionador de PNOF

def PNOFi_selector(n,p):
    if(p.ipnof==5):
        cj12,ck12 = CJCKD5(n,p.no1,p.ndoc,p.nsoc,p.nbeta,p.nalpha,p.ndns,p.ncwo,p.MSpin)
    if(p.ipnof==7):
        cj12,ck12 = CJCKD7(n,p.ista,p.no1,p.ndoc,p.nsoc,p.nbeta,p.nalpha,p.ndns,p.ncwo,p.MSpin)
    if(p.ipnof==8):
        cj12,ck12 = CJCKD8(n,p.no1,p.ndoc,p.nsoc,p.nbeta,p.nalpha,p.ndns,p.ncwo,p.MSpin)

    return cj12,ck12

def der_PNOFi_selector(n,dn_dgamma,p):
    if(p.ipnof==5):
        Dcj12r,Dck12r = der_CJCKD5(n,p.ista,dn_dgamma,p.no1,p.ndoc,p.nalpha,p.nbeta,p.nv,p.nbf5,p.ndns,p.ncwo)
    if(p.ipnof==7):
        Dcj12r,Dck12r = der_CJCKD7(n,p.ista,dn_dgamma,p.no1,p.ndoc,p.nalpha,p.nbeta,p.nv,p.nbf5,p.ndns,p.ncwo)
    if(p.ipnof==8):
        Dcj12r,Dck12r = der_CJCKD8(n,dn_dgamma,p.no1,p.ndoc,p.nalpha,p.nbeta,p.nv,p.nbf5,p.ndns,p.ncwo,p.MSpin,p.nsoc)
        
    return Dcj12r,Dck12r

@njit(cache=True)
def ocupacion(gamma,no1,ndoc,nalpha,nv,nbf5,ndns,ncwo,HighSpin,occ_method):
    if occ_method == "Trigonometric":
        n,dn_dgamma = ocupacion_trigonometric(gamma,no1,ndoc,nalpha,nv,nbf5,ndns,ncwo,HighSpin)
    elif occ_method == "Softmax":
        n,dn_dgamma = ocupacion_softmax(gamma,no1,ndoc,nalpha,nv,nbf5,ndns,ncwo,HighSpin)
    return n,dn_dgamma

@njit(cache=True)
def ocupacion_trigonometric(gamma,no1,ndoc,nalpha,nv,nbf5,ndns,ncwo,HighSpin):

    n = np.zeros((nbf5))
    dni_dgammai = np.zeros((nbf5))

    n[0:no1] = 1                                              # [1,no1]

    n[no1:no1+ndoc] = 1/2 * (1 + np.cos(gamma[:ndoc])**2)     # (no1,no1+ndoc]
    dni_dgammai[no1:no1+ndoc] = - 1/2 * np.sin(2*gamma[:ndoc])
    
    if(not HighSpin):
        n[no1+ndoc:no1+ndns] = 0.5   # (no1+ndoc,no1+ndns]
    elif(HighSpin):
        n[no1+ndoc:no1+ndns] = 1.0   # (no1+ndoc,no1+ndns]

    if(ncwo==1):
        dn_dgamma = np.zeros((nbf5,nv))

        for i in range(ndoc):
            dn_dgamma[no1+i][i] = dni_dgammai[no1+i]
            #cwo
            icf = nalpha + ndoc - i - 1
            n[icf] = 1/2*np.sin(gamma[i])**2
            dni_dgammai[icf]  = 1/2*np.sin(2*gamma[i])
            dn_dgamma[icf][i] = dni_dgammai[icf]
    else:
        dn_dgamma = np.zeros((nbf5,nv))
        h = 1 - n
        for i in range(ndoc):
            ll = no1 + ndns + ncwo*(ndoc - i - 1)
            ul = no1 + ndns + ncwo*(ndoc - i)
            n[ll:ul] = h[no1+i]
            for iw in range(ncwo-1):
                n[ll+iw] *= np.sin(gamma[ndoc+(ncwo-1)*i+iw])**2
                n[ll+iw+1:ul] *= np.cos(gamma[ndoc+(ncwo-1)*i+iw])**2

        for i in range(ndoc):
            # dn_g/dgamma_g
            dn_dgamma[no1+i][i] = dni_dgammai[no1+i]

            # dn_pi/dgamma_g
            ll = no1 + ndns + ncwo*(ndoc - i - 1)
            ul = no1 + ndns + ncwo*(ndoc - i)
            dn_dgamma[ll:ul,i] = -dni_dgammai[no1+i]
            for iw in range(ncwo-1):
                dn_dgamma[ll+iw][i] *= np.sin(gamma[ndoc+(ncwo-1)*i+iw])**2
                dn_dgamma[ll+iw+1:ul,i] *= np.cos(gamma[ndoc+(ncwo-1)*i+iw])**2

            # dn_pi/dgamma_pj (j<i)
            for iw in range(ncwo-1):
                dn_dgamma[ll+iw+1:ul,ndoc+(ncwo-1)*i+iw] = n[no1+i] - 1
                for ip in range(ll+iw+1,ul):
                    for jw in range(ip-ll):
                        if(jw==iw):
                            dn_dgamma[ip][ndoc+(ncwo-1)*i+iw] *= np.sin(2*gamma[ndoc+(ncwo-1)*i+jw])  
                        else:
                            dn_dgamma[ip][ndoc+(ncwo-1)*i+iw] *= np.cos(gamma[ndoc+(ncwo-1)*i+jw])**2  
                    if(ip-ll<ncwo-1):
                        dn_dgamma[ip][ndoc+(ncwo-1)*i+iw] *= np.sin(gamma[ndoc+(ncwo-1)*i+(ip-ll)])**2  

            # dn_pi/dgamma_i
            for iw in range(ncwo-1):
                dn_dgamma[ll+iw][ndoc+(ncwo-1)*i+iw] = 1 - n[no1+i]
                for jw in range(iw+1):
                    if(jw==iw):
                        dn_dgamma[ll+iw][ndoc+(ncwo-1)*i+iw] *= np.sin(2*gamma[ndoc+(ncwo-1)*i+jw])  
                    else:
                        dn_dgamma[ll+iw][ndoc+(ncwo-1)*i+iw] *= np.cos(gamma[ndoc+(ncwo-1)*i+jw])**2  

    return n,dn_dgamma

@njit(cache=True)
def ocupacion_softmax(x,no1,ndoc,nalpha,nv,nbf5,ndns,ncwo,HighSpin):

    n = np.zeros((nbf5))
    dn_dx = np.zeros((nbf5,nv))

    if(not HighSpin):
         n[no1+ndoc:no1+ndns] = 0.5   # (no1+ndoc,no1+ndns]
    elif(HighSpin):
         n[no1+ndoc:no1+ndns] = 1.0   # (no1+ndoc,no1+ndns]

    exp_x = np.exp(x)

    for i in range(ndoc):
        ll = no1 + ndns + ncwo*(ndoc - i - 1)
        ul = no1 + ndns + ncwo*(ndoc - i)

        ll_x = ll - ndns + ndoc - no1
        ul_x = ul - ndns + ndoc - no1

        sum_exp = exp_x[i] + np.sum(exp_x[ll_x:ul_x])

        n[i] = exp_x[i]/sum_exp
        n[ll:ul] = exp_x[ll_x:ul_x]/sum_exp

        dn_dx[ll:ul,ll_x:ul_x] = -np.outer(exp_x[ll_x:ul_x],exp_x[ll_x:ul_x])/sum_exp**2

        dn_dx[i,ll_x:ul_x] = -exp_x[ll_x:ul_x]*exp_x[i]/sum_exp**2
        dn_dx[ll:ul,i] = -exp_x[ll_x:ul_x]*exp_x[i]/sum_exp**2

        dn_dx[i,i] = exp_x[i]*(sum_exp-exp_x[i])/sum_exp**2

        for j in range(ncwo):
            dn_dx[ll+j,ll_x+j] = exp_x[ll_x+j]*(sum_exp-exp_x[ll_x+j])/sum_exp**2

    return n,dn_dx

def calce(n,cj12,ck12,J_MO,K_MO,H_core,p):

    E = 0

    if(p.MSpin==0):

        # 2H + J
        E = E + 2*np.einsum('i,i',n,H_core,optimize=True)
        E = E + np.einsum('i,i',n[:p.nbeta],np.diagonal(J_MO)[:p.nbeta],optimize=True)
        E = E + np.einsum('i,i',n[p.nalpha:p.nbf5],np.diagonal(J_MO)[p.nalpha:p.nbf5],optimize=True)
    
        #C^J JMO
        np.fill_diagonal(cj12,0) # Remove diag.
        E = E + np.einsum('ij,ji->',cj12,J_MO,optimize=True) # sum_ij
    
        #C^K KMO
        np.fill_diagonal(ck12,0) # Remove diag.
        E = E - np.einsum('ij,ji->',ck12,K_MO,optimize=True) # sum_ij
    
    elif(not p.MSpin==0):
        E = 0

        # 2H + J
        E = E + np.einsum('i,i',n[:p.nbeta],2*H_core[:p.nbeta]+np.diagonal(J_MO)[:p.nbeta],optimize=True) # [0,Nbeta]
        E = E + np.einsum('i,i',n[p.nbeta:p.nalpha],2*H_core[p.nbeta:p.nalpha],optimize=True)               # (Nbeta,Nalpha]
        E = E + np.einsum('i,i',n[p.nalpha:p.nbf5],2*H_core[p.nalpha:p.nbf5]+np.diagonal(J_MO)[p.nalpha:p.nbf5],optimize=True) # (Nalpha,Nbf5)

        #C^J JMO
        np.fill_diagonal(cj12,0) # Remove diag.
        E = E + np.einsum('ij,ji->',cj12[:p.nbeta,:p.nbeta],J_MO[:p.nbeta,:p.nbeta],optimize=True) # sum_ij
        E = E + np.einsum('ij,ji->',cj12[:p.nbeta,p.nalpha:p.nbf5],J_MO[p.nalpha:p.nbf5,:p.nbeta],optimize=True) # sum_ij
        E = E + np.einsum('ij,ji->',cj12[p.nalpha:p.nbf5,:p.nbeta],J_MO[:p.nbeta,p.nalpha:p.nbf5],optimize=True) # sum_ij
        E = E + np.einsum('ij,ji->',cj12[p.nalpha:p.nbf5,p.nalpha:p.nbf5],J_MO[p.nalpha:p.nbf5,p.nalpha:p.nbf5],optimize=True) # sum_ij

        #C^K KMO
        np.fill_diagonal(ck12,0) # Remove diag.
        E = E - np.einsum('ij,ji->',ck12[:p.nbeta,:p.nbeta],K_MO[:p.nbeta,:p.nbeta],optimize=True) # sum_ij
        E = E - np.einsum('ij,ji->',ck12[:p.nbeta,p.nalpha:p.nbf5],K_MO[p.nalpha:p.nbf5,:p.nbeta],optimize=True) # sum_ij
        E = E - np.einsum('ij,ji->',ck12[p.nalpha:p.nbf5,:p.nbeta],K_MO[:p.nbeta,p.nalpha:p.nbf5],optimize=True) # sum_ij
        E = E - np.einsum('ij,ji->',ck12[p.nalpha:p.nbf5,p.nalpha:p.nbf5],K_MO[p.nalpha:p.nbf5,p.nalpha:p.nbf5],optimize=True) # sum_ij

        #n JMO
        E = E + 2*np.einsum('i,ji->',n[:p.nbeta],J_MO[p.nbeta:p.nalpha,:p.nbeta],optimize=True) # sum_ij
        E = E + 2*np.einsum('i,ji->',n[p.nalpha:p.nbf5],J_MO[p.nbeta:p.nalpha,p.nalpha:p.nbf5],optimize=True) # sum_ij
        E = E + 0.5*(np.einsum('i,ji->',n[p.nbeta:p.nalpha],J_MO[p.nbeta:p.nalpha,p.nbeta:p.nalpha],optimize=True) - np.einsum('i,ii->',n[p.nbeta:p.nalpha],J_MO[p.nbeta:p.nalpha,p.nbeta:p.nalpha],optimize=True))

        #n KMO
        E = E - np.einsum('i,ji->',n[:p.nbeta],K_MO[p.nbeta:p.nalpha,:p.nbeta],optimize=True) # sum_ij
        E = E - np.einsum('i,ji->',n[p.nalpha:p.nbf5],K_MO[p.nbeta:p.nalpha,p.nalpha:p.nbf5],optimize=True) # sum_ij
        E = E - 0.5*(np.einsum('i,ji->',n[p.nbeta:p.nalpha],K_MO[p.nbeta:p.nalpha,p.nbeta:p.nalpha],optimize=True) + np.einsum('i,ii->',n[p.nbeta:p.nalpha],K_MO[p.nbeta:p.nalpha,p.nbeta:p.nalpha],optimize=True))

    return E

def calcocce(gamma,J_MO,K_MO,H_core,p):

    n,dn_dgamma = ocupacion(gamma,p.no1,p.ndoc,p.nalpha,p.nv,p.nbf5,p.ndns,p.ncwo,p.HighSpin,p.occ_method)
    cj12,ck12 = PNOFi_selector(n,p)

    E = calce(n,cj12,ck12,J_MO,K_MO,H_core,p)

    return E

def calcoccg(gamma,J_MO,K_MO,H_core,p):

    grad = np.zeros((p.nv))

    n,dn_dgamma = ocupacion(gamma,p.no1,p.ndoc,p.nalpha,p.nv,p.nbf5,p.ndns,p.ncwo,p.HighSpin,p.occ_method)
    Dcj12r,Dck12r = der_PNOFi_selector(n,dn_dgamma,p)

    if(p.MSpin==0):

        # dn_dgamma (2H+J)
        grad += np.einsum('ik,i->k',dn_dgamma,2*H_core+np.diagonal(J_MO),optimize=True)

        # 2 dCJ_dgamma J_MO
        diag = np.diag_indices(p.nbf5)
        Dcj12r[diag] = 0
        grad += 2*np.einsum('ijk,ji->k',Dcj12r,J_MO,optimize=True)
    
        # -2 dCK_dgamma K_MO
        diag = np.diag_indices(p.nbf5)
        Dck12r[diag] = 0
        grad -= 2*np.einsum('ijk,ji->k',Dck12r,K_MO,optimize=True)

    elif(not p.MSpin==0):
    
        # dn_dgamma (2H+J)
        grad += np.einsum('ik,i->k',dn_dgamma[p.no1:p.nbeta,:p.nv],2*H_core[p.no1:p.nbeta]+np.diagonal(J_MO)[p.no1:p.nbeta],optimize=True) # [0,Nbeta]
        grad += np.einsum('ik,i->k',dn_dgamma[p.nalpha:p.nbf5,:p.nv],2*H_core[p.nalpha:p.nbf5]+np.diagonal(J_MO)[p.nalpha:p.nbf5],optimize=True) # [Nalpha,Nbf5]

        # 2 dCJ_dgamma J_MO
        Dcj12r[np.diag_indices(p.nbf5)] = 0
        grad += 2*np.einsum('ijk,ji->k',Dcj12r[p.no1:p.nbeta,:p.nbeta,:p.nv],J_MO[:p.nbeta,p.no1:p.nbeta],optimize=True)
        grad += 2*np.einsum('ijk,ji->k',Dcj12r[p.no1:p.nbeta,p.nalpha:p.nbf5,:p.nv],J_MO[p.nalpha:p.nbf5,p.no1:p.nbeta],optimize=True)
        grad += 2*np.einsum('ijk,ji->k',Dcj12r[p.nalpha:p.nbf5,:p.nbeta,:p.nv],J_MO[:p.nbeta,p.nalpha:p.nbf5],optimize=True)
        grad += 2*np.einsum('ijk,ji->k',Dcj12r[p.nalpha:p.nbf5,p.nalpha:p.nbf5,:p.nv],J_MO[p.nalpha:p.nbf5,p.nalpha:p.nbf5],optimize=True)

        # -2 dCK_dgamma K_MO
        Dck12r[np.diag_indices(p.nbf5)] = 0
        grad -= 2*np.einsum('ijk,ji->k',Dck12r[p.no1:p.nbeta,:p.nbeta,:p.nv],K_MO[:p.nbeta,p.no1:p.nbeta],optimize=True)
        grad -= 2*np.einsum('ijk,ji->k',Dck12r[p.no1:p.nbeta,p.nalpha:p.nbf5,:p.nv],K_MO[p.nalpha:p.nbf5,p.no1:p.nbeta],optimize=True)
        grad -= 2*np.einsum('ijk,ji->k',Dck12r[p.nalpha:p.nbf5,:p.nbeta,:p.nv],K_MO[:p.nbeta,p.nalpha:p.nbf5],optimize=True)
        grad -= 2*np.einsum('ijk,ji->k',Dck12r[p.nalpha:p.nbf5,p.nalpha:p.nbf5,:p.nv],K_MO[p.nalpha:p.nbf5,p.nalpha:p.nbf5],optimize=True)

        # 2 dn_dgamma J_MO
        grad += 2*np.einsum('jk,ij->k',dn_dgamma[p.no1:p.nbeta,:p.nv],J_MO[p.nbeta:p.nalpha,p.no1:p.nbeta],optimize=True)
        grad += 2*np.einsum('jk,ij->k',dn_dgamma[p.nalpha:p.nbf5,:p.nv],J_MO[p.nbeta:p.nalpha,p.nalpha:p.nbf5],optimize=True)

        # - dn_dgamma K_MO
        grad -= np.einsum('jk,ij->k',dn_dgamma[p.no1:p.nbeta,:p.nv],K_MO[p.nbeta:p.nalpha,p.no1:p.nbeta],optimize=True)
        grad -= np.einsum('jk,ij->k',dn_dgamma[p.nalpha:p.nbf5,:p.nv],K_MO[p.nbeta:p.nalpha,p.nalpha:p.nbf5],optimize=True)

    return grad

def calcorbe(y,n,cj12,ck12,C,H,I,b_mnl,p):

    Cnew = pynof.rotate_orbital(y,C,p)

    J_MO,K_MO,H_core = pynof.computeJKH_MO(Cnew,H,I,b_mnl,p)

    E = calce(n,cj12,ck12,J_MO,K_MO,H_core,p)

    return E

def calcorbg(y,n,cj12,ck12,C,H,I,b_mnl,p):

    Cnew = pynof.rotate_orbital(y,C,p)

    elag,Hmat = pynof.computeLagrange2(n,cj12,ck12,Cnew,H,I,b_mnl,p)

    grad = 4*(elag - elag.T)

    grads = np.zeros((int(p.nbf*(p.nbf-1)/2) - int(p.no0*(p.no0-1)/2)))

    k = 0
    for i in range(p.nbf5):
        for j in range(i+1,p.nbf):
            grads[k] = grad[i,j]
            k += 1
    grad = grads

    return grad

def calcorbeg(y,n,cj12,ck12,C,H,I,b_mnl,p):

    Cnew = pynof.rotate_orbital(y,C,p)

    elag,Hmat = pynof.computeLagrange2(n,cj12,ck12,Cnew,H,I,b_mnl,p)

    grad = 4*(elag - elag.T)

    grads = np.zeros((int(p.nbf*(p.nbf-1)/2) - int(p.no0*(p.no0-1)/2)))

    k = 0
    for i in range(p.nbf5):
        for j in range(i+1,p.nbf):
            grads[k] = grad[i,j]
            k += 1
    grad = grads

    E = np.einsum('ii',elag[:p.nbf5,:p.nbf5],optimize=True)
    E = E + np.einsum('i,ii',n[:p.nbf5],Hmat[:p.nbf5,:p.nbf5],optimize=True)

    return E,grad



import psi4
from scipy.linalg import orth
from numpy.linalg import matrix_power
import scipy
import numpy as np
import integrals
from scipy.linalg import eigh
import utils
import minimization
from numpy.linalg import lstsq
from scipy.linalg import orth

def find_orth(O,vec):
    A = np.hstack((O, vec.reshape((O.shape[0],1))))
    b = np.zeros(O.shape[1] + 1)
    b[-1] = 1
    return lstsq(A.T, b, rcond=None)[0]

def project_MO(mol,C,old_basis):

    wfn = psi4.core.Wavefunction.build(mol, psi4.core.get_global_option('basis'))
    mints = psi4.core.MintsHelper(wfn.basisset())

    M = np.asarray(mints.ao_overlap(wfn.basisset(),old_basis))

    Sinv = mints.ao_overlap()
    Sm12 = mints.ao_overlap()
    Sp12 = mints.ao_overlap()
    Sinv.power(-1.0, 1.e-14)
    Sm12.power(-0.5, 1.e-14)
    Sp12.power(0.5, 1.e-14)

    Cnew = np.matmul(np.matmul(Sinv,M),C)


    # Orthonormalize in C' = S^1/2 * C
    Cp = np.matmul(Sp12,Cnew)

    Cpnew = Cp[:,0]
    Cpnew = Cpnew.reshape((Cpnew.shape[0],1))

    for i in range(1,Cp.shape[1]):
        res = find_orth(Cpnew,Cp[:,i])
        res = res/np.linalg.norm(res)

        Cpnew = np.hstack((Cpnew,res.reshape((Cpnew.shape[0],1))))

    C = np.matmul(Sm12,Cpnew)    

    return Cnew

def complete_projection(wfn,mol,p,C,fmiug0,printmode=False):

    S,T,V,H,I,b_mnl = integrals.compute_integrals(wfn,mol,p)
    mints = psi4.core.MintsHelper(wfn.basisset())
    Sm12 = mints.ao_overlap()
    Sm12.power(-0.5, 1.e-14)
    Sp12 = mints.ao_overlap()
    Sp12.power(0.5, 1.e-14)

    # Energía Nuclear
    E_nuc = mol.nuclear_repulsion_energy()

    # Guess de MO (C)
    E_i,Cguess = eigh(H, S)  # (HC = SCe)
    Cguess = utils.check_ortho(Cguess,S,p)

    # HF MO
    if (p.hfidr):
        EHF,Cguess,fmiug0guess = minimization.hfidr(Cguess,H,I,b_mnl,E_nuc,p,printmode=True)

    # Combine projected MO
    Cguess[:,:C.shape[1]] = C[:,:C.shape[1]]

    # Orthonormalize in C' = S^1/2 * C
    Cp = np.matmul(Sp12,Cguess)

    Cpnew = Cp[:,0]
    Cpnew = Cpnew.reshape((Cpnew.shape[0],1))

    for i in range(1,Cp.shape[1]):
        res = find_orth(Cpnew,Cp[:,i])
        res = res/np.linalg.norm(res)

        Cpnew = np.hstack((Cpnew,res.reshape((Cpnew.shape[0],1))))

    C = np.matmul(Sm12,Cpnew)    

    C = utils.check_ortho(C,S,p)

    fmiug0guess[:fmiug0.shape[0]] = fmiug0
    fmiug0 = fmiug0guess

    return C,fmiug0



import numpy as np
import psi4

class param():
   
    def __init__(self,mol,basname):

        psi4.set_options({'basis': basname})

        # Paramdetros del sistema
        self.wfn = wfn = psi4.core.Wavefunction.build(mol, psi4.core.get_global_option('basis'))

        # Parámetros 
        self.natoms = mol.natom()
        self.nbf = wfn.basisset().nbf()
        self.nbfaux = 0
        self.nalpha = wfn.nalpha()
        self.nbeta = wfn.nbeta()
        self.ne = self.nalpha + self.nbeta
        self.mul = mol.multiplicity()
        self.charge = mol.molecular_charge()
        no1 = 0 #Number of inactive doubly occupied orbitals | Se puede variar
        #for i in range(self.natoms):
        #    Z = mol.Z(i)
        #    if ( 1<=Z and Z<=  2):
        #        no1 += 0           # H-He
        #    elif ( 3<=Z and Z<= 10):
        #        no1 +=  1          # Li-Ne
        #    elif (11<=Z and Z<= 18):
        #        no1 +=  5          # Na-Ar
        #    elif(19<=Z and Z<= 36):
        #        no1 +=  9          # K-Kr
        #    elif(37<=Z and Z<= 49):
        #        no1 += 18          # Rb-In
        #    elif(50<=Z and Z<= 54):
        #        no1 += 23          # Sn-Xe
        #    elif(55<=Z and Z<= 71):
        #        no1 += 27          # Cs-Lu
        #    elif(72<=Z and Z<= 81):
        #        no1 += 30          # Hf-Tl
        #    elif(82<=Z and Z<= 86):
        #        no1 += 39          # Pb-Rn
        #    elif(87<=Z and Z<=109):
        #        no1 += 43          # Fr-Mt
        self.ndoc = self.nbeta   -   no1
        self.nsoc = self.nalpha  -   self.nbeta
        self.ndns = self.ndoc    +   self.nsoc
        self.nvir = self.nbf     -   self.nalpha
        
        ncwo = -1
        if(self.ne==2):
            ncwo= -1
        if(self.ndns!=0):
            if(self.ndoc>0):
                if(ncwo!=1):
                    if(ncwo==-1 or ncwo > self.nvir/self.ndoc):
                        ncwo = int(self.nvir/self.ndoc)
            else:
                ncwo = 0

        self.closed = (self.nbeta == (self.ne+self.mul-1)/2 and self.nalpha == (self.ne-self.mul+1)/2)
        
        self.nac = self.ndoc * (1 + ncwo)
        self.nbf5 = no1 + self.nac + self.nsoc   #JFHLY warning: nbf must be >nbf5
        self.no0 = self.nbf - self.nbf5

        noptorb = self.nbf
  
        self.title = "pynof"
        self.maxit = 1000  # Número máximo de iteraciones de Occ-SCF
        self.no1 = no1     # Número de orbitales inactivos con ocupación 1
        self.thresheid = 10**-6#8 # Convergencia de la energía total
        self.maxitid = 30  # Número máximo de iteraciones externas en HF
        self.maxloop = 30  # Iteraciones internas en optimización orbital
        self.ipnof = 8     # PNOFi a calcular
        self.ista = 0     # PNOFi a calcular
        self.threshl = 10**-3   # Convergencia de los multiplicadores de Lagrange
        self.threshe = 10**-4   # Convergencia de la energía
        self.threshec = 10**-8 # Convergencia  de la energía en optimización orbital
        self.threshen = 10**-10 # Convergencia  de la energía en optimización de ocupaciones
        self.scaling = True     # Scaling for f
        self.nzeros = 0
        self.nzerosm = 5
        self.nzerosr = 2
        self.itziter = 10        # Iteraciones para scaling constante
        self.diis = True         # DIIS en optimización orbital
        self.thdiis = 10**-3     # Para iniciar DIIS
        self.ndiis = 5           # Número de ciclos para interpolar matriz de Fock generalizada en DIIS
        self.perdiis = True      # Aplica DIIS cada NDIIS (True) o después de NDIIS (False)
        self.ncwo = ncwo         # Número de orbitales débilmente ocupados acoplados a cada orbital fueremtente ocupado
        self.noptorb = noptorb   # Número de orbitales a optimizar Nbf5 <= Noptorb <= Nbf
        self.scaling = True
        self.nv = self.ncwo*self.ndoc
        self.nvar = int(self.nbf*(self.nbf-1)/2) - int(self.no0*(self.no0-1)/2)
        self.occupation_optimizer = "CG"
        self.orbital_optimizer = "CG"
        self.combined_optimizer = "CG"
        self.tol_gorb = 1e-3
        self.tol_gocc = 1e-2
        self.gpu = False
        self.RI = False
        self.jit= False

        self.HighSpin = False
        self.MSpin = 0

        self.lamb = 0.0
        self.occ_method = "Softmax"
        self.orb_method = "ID"

    def autozeros(self,restart=False):
        if(restart):
            if(abs(np.log10(self.threshl))<=3):
                self.nzeros = 2
                self.nzerosr = 2
                self.nzerosm = 5
            else:
                self.nzeros = abs(int(np.log10(self.threshl))) - 1
                self.nzerosr = self.nzeros
                self.nzerosm = abs(int(np.log10(self.threshl))) + 2         
        else:
            #if(abs(np.log10(self.threshl))>3):
            self.nzeros = 1
            self.nzerosr = 2
            self.nzerosm = abs(int(np.log10(self.threshl))) + 2

    def set_ncwo(self,ncwo):
        #if(self.ne==2):
        #    ncwo= -1
        if(self.ndns!=0):
            if(self.ndoc>0):
                if(ncwo!=1):
                    if(ncwo==-1 or ncwo > self.nvir/self.ndoc):
                        ncwo = int(self.nvir/self.ndoc)
            else:
                ncwo = 0

        self.ncwo = ncwo

        self.nac = self.ndoc * (1 + ncwo)
        self.nbf5 = self.no1 + self.nac + self.nsoc   #JFHLY warning: nbf must be >nbf5
        self.no0 = self.nbf - self.nbf5
        self.nvar = int(self.nbf*(self.nbf-1)/2) - int(self.no0*(self.no0-1)/2)
        self.nv = self.ncwo*self.ndoc

    def remove_no1(self,gamma=None):

        no1_old = self.no1

        no1 = 0 #Number of inactive doubly occupied orbitals | Se puede variar
        self.ndoc = self.nbeta   -   no1
        self.ndns = self.ndoc    +   self.nsoc

        self.nac = self.ndoc * (1 + self.ncwo)
        self.nbf5 = no1 + self.nac + self.nsoc   #JFHLY warning: nbf must be >nbf5
        self.no0 = self.nbf - self.nbf5

        self.no1 = no1     # Número de orbitales inactivos con ocupación 1
        self.nv = self.ncwo*self.ndoc
        self.nvar = int(self.nbf*(self.nbf-1)/2) - int(self.no0*(self.no0-1)/2)

        if(not gamma is None):
            gamma = np.concatenate((np.ones((no1_old))*np.arccos(np.sqrt(2.0*0.999-1.0)),gamma))
            return gamma





import numpy as np

def fchk(filename,wfn,mol,jobtype,E_t,elag,n,C,p):

    C_new = C.copy()

    b = wfn.basisset()
    if(b.has_puream()):
        for ishell in range(b.nshell()):
            shell = b.shell(ishell)
            if(shell.am==1):
                idx = b.shell_to_basis_function(ishell)
                tmp_coeffs = C_new[idx:idx+3,:].copy()
                C_new[idx+0,:] = tmp_coeffs[1,:]
                C_new[idx+1,:] = tmp_coeffs[2,:]
                C_new[idx+2,:] = tmp_coeffs[0,:]
    elif(not b.has_puream()):
        for ishell in range(b.nshell()):
            shell = b.shell(ishell)
            if(shell.am==2):
                idx = b.shell_to_basis_function(ishell)
                tmp_coeffs = C_new[idx:idx+6,:].copy()
                C_new[idx+0,:] = tmp_coeffs[0,:]
                C_new[idx+1,:] = tmp_coeffs[3,:]
                C_new[idx+2,:] = tmp_coeffs[5,:]
                C_new[idx+3,:] = tmp_coeffs[1,:]/np.sqrt(3)
                C_new[idx+4,:] = tmp_coeffs[2,:]/np.sqrt(3)
                C_new[idx+5,:] = tmp_coeffs[4,:]/np.sqrt(3)
            elif(shell.am==3):
                idx = b.shell_to_basis_function(ishell)
                tmp_coeffs = C_new[idx:idx+10,:].copy()
                C_new[idx+0,:] = tmp_coeffs[0,:]
                C_new[idx+1,:] = tmp_coeffs[6,:]
                C_new[idx+2,:] = tmp_coeffs[9,:]
                C_new[idx+3,:] = tmp_coeffs[3,:]/np.sqrt(5)
                C_new[idx+4,:] = tmp_coeffs[1,:]/np.sqrt(5)
                C_new[idx+5,:] = tmp_coeffs[2,:]/np.sqrt(5)
                C_new[idx+6,:] = tmp_coeffs[5,:]/np.sqrt(5)
                C_new[idx+7,:] = tmp_coeffs[8,:]/np.sqrt(5)
                C_new[idx+8,:] = tmp_coeffs[7,:]/np.sqrt(5)
                C_new[idx+9,:] = tmp_coeffs[4,:]/np.sqrt(15)
            elif(shell.am==4):
                idx = b.shell_to_basis_function(ishell)
                tmp_coeffs = C_new[idx:idx+15,:].copy()
                C_new[idx+0,:] = tmp_coeffs[0,:]
                C_new[idx+1,:] = tmp_coeffs[10,:]
                C_new[idx+2,:] = tmp_coeffs[14,:]
                C_new[idx+3,:] = tmp_coeffs[1,:]/np.sqrt(7)
                C_new[idx+4,:] = tmp_coeffs[2,:]/np.sqrt(7)
                C_new[idx+5,:] = tmp_coeffs[6,:]/np.sqrt(7)
                C_new[idx+6,:] = tmp_coeffs[11,:]/np.sqrt(7)
                C_new[idx+7,:] = tmp_coeffs[9,:]/np.sqrt(7)
                C_new[idx+8,:] = tmp_coeffs[13,:]/np.sqrt(7)
                C_new[idx+9,:] = tmp_coeffs[3,:]*np.sqrt(3)/np.sqrt(35)
                C_new[idx+10,:] = tmp_coeffs[5,:]*np.sqrt(3)/np.sqrt(35)
                C_new[idx+11,:] = tmp_coeffs[12,:]*np.sqrt(3)/np.sqrt(35)
                C_new[idx+12,:] = tmp_coeffs[4,:]/np.sqrt(35)
                C_new[idx+13,:] = tmp_coeffs[7,:]/np.sqrt(35)
                C_new[idx+14,:] = tmp_coeffs[8,:]/np.sqrt(35)
            elif(shell.am>=5):
                print("Warning, angular momentum not supported in fchk")


    f = open(filename+".fchk","w")

    print("{}".format(filename),file=f)
    if(p.ista==0):
        print("{} PNOF{} {}".format(jobtype,p.ipnof,wfn.basisset().blend()),file=f)
    else:
        print("{} PNOF{}s {}".format(jobtype,p.ipnof,wfn.basisset().blend()),file=f)
    print("Number of atoms                            I           {:6d}".format(p.natoms),file=f)
    print("Charge                                     I           {:6d}".format(p.charge),file=f)
    print("Multiplicity                               I           {:6d}".format(p.mul),file=f)
    print("Number of electrons                        I           {:6d}".format(p.ne),file=f)
    print("Number of alpha electrons                  I           {:6d}".format(p.nalpha),file=f)
    print("Number of beta electrons                   I           {:6d}".format(p.nbeta),file=f)
    print("Number of basis functions                  I           {:6d}".format(p.nbf),file=f)
    print("Number of independant functions            I           {:6d}".format(p.nbf5),file=f)
    print("Number of contracted shells                I           {:6d}".format(wfn.basisset().nshell()),file=f)
    print("Highest angular momentum                   I           {:6d}".format(wfn.basisset().max_am()),file=f)
    print("Largest degree of contraction              I           {:6d}".format(wfn.basisset().max_nprimitive()),file=f)
    print("Number of primitive shells                 I           {:6d}".format(wfn.basisset().nprimitive()),file=f)
#    print("Virial Ratio                               R                {}".format(p.natoms),file=f)
#    print("SCF Energy                                 R                {}".format(p.natoms),file=f)
    print("Atomic numbers                             I   N=      {:6d}".format(p.natoms),file=f)
    for i in range(p.natoms):
        Z = mol.Z(i)
        print(" {:11d}".format(int(Z)),end="",file=f)
        if((i+1)%6==0 or i+1==p.natoms):
                print("",file=f)
    print("Nuclear Charges                            R   N=      {:6d}".format(p.natoms),file=f)
    for i in range(p.natoms):
        Z = mol.Z(i)
        print(" {: .8e}".format(Z),end="",file=f)
        if((i+1)%6==0 or i+1==p.natoms):
                print("",file=f)
    print("Current cartesian coordinates              R   N=      {:6d}".format(p.natoms*3),file=f)
    coord, mass, symbols, Z, key = wfn.molecule().to_arrays()
    idata = 0
    for xyz in coord:
        for ixyz in range(3):
            idata += 1
            print(" {: .8e}".format(xyz[ixyz]),end="",file=f)
            if(idata%5==0 or idata==p.natoms*3):
                print("",file=f)
    print("Shell types                                I   N=      {:6d}".format(wfn.basisset().nshell()),file=f)
    for ishell in range(wfn.basisset().nshell()):
        if(wfn.basisset().has_puream() and wfn.basisset().shell(ishell).am > 1):
            print(" {:11d}".format(-1*wfn.basisset().shell(ishell).am),end ="", file=f)
        else:
            print(" {:11d}".format(wfn.basisset().shell(ishell).am),end ="", file=f)
        if((ishell+1)%6==0 or ishell+1==wfn.basisset().nshell()):
            print("",file=f)
    print("Number of primitives per shell             I   N=      {:6d}".format(wfn.basisset().nshell()),file=f)
    for ishell in range(wfn.basisset().nshell()):
        print(" {:11d}".format(wfn.basisset().shell(ishell).nprimitive),end ="", file=f)
        if((ishell+1)%6==0 or ishell+1==wfn.basisset().nshell()):
            print("",file=f)
    print("Shell to atom map                          I   N=      {:6d}".format(wfn.basisset().nshell()),file=f)
    for ishell in range(wfn.basisset().nshell()):
        print(" {:11d}".format(wfn.basisset().shell(ishell).ncenter+1),end ="", file=f)
        if((ishell+1)%6==0 or ishell+1==wfn.basisset().nshell()):
            print("",file=f)
    print("Coordinates of each shell                  R   N=      {:6d}".format(wfn.basisset().nshell()*3),file=f)
    idata = 0
    for ishell in range(wfn.basisset().nshell()):
        xyz = coord[wfn.basisset().shell(ishell).ncenter]
        for ixyz in range(3):
            idata += 1
            print(" {: .8e}".format(xyz[ixyz]),end ="", file=f)
            if(idata%5==0 or idata==wfn.basisset().nshell()*3):
                print("",file=f)
    print("Total Energy                               R     {: .15e}".format(E_t),file=f)
    print("Primitive exponents                        R   N=      {:6d}".format(wfn.basisset().nprimitive()),file=f)
    idata = 0
    for ishell in range(wfn.basisset().nshell()):
        for iprim in range(wfn.basisset().shell(ishell).nprimitive):
            idata += 1
            print(" {: .8e}".format(wfn.basisset().shell(ishell).exp(iprim)),end ="", file=f)
            if(idata%5==0 or idata==wfn.basisset().nprimitive()):
                print("",file=f)
    print("Contraction coefficients                   R   N=      {:6d}".format(wfn.basisset().nprimitive()),file=f)
    idata = 0
    for ishell in range(wfn.basisset().nshell()):
        for iprim in range(wfn.basisset().shell(ishell).nprimitive):
            idata += 1
            print(" {: .8e}".format(wfn.basisset().shell(ishell).original_coef(iprim)),end ="", file=f)
            if(idata%5==0 or idata==wfn.basisset().nprimitive()):
                print("",file=f)

    e_val = elag[np.diag_indices(p.nbf)]

    print("Alpha Orbital Energies                     R   N=      {:6d}".format(p.nbf),file=f)
    for i in range(p.nbf):
        print(" {: .8e}".format(e_val[i]),end ="", file=f)
        if((i+1)%5==0 or i+1==p.nbf):
            print("",file=f)
#    print("Beta Orbital Energies                      R   N=      {:6d}".format(p.nbf),file=f)
#    for i in range(p.nbf):
#        print(" {: .8e}".format(e_val[i]),end ="", file=f)
#        if((i+1)%5==0 or i+1==p.nbf):
#            print("",file=f)
    print("Alpha MO coefficients                      R   N=      {:6d}".format(p.nbf*p.nbf),file=f)
    idata = 0
    for j in range(p.nbf):
        for i in range(p.nbf):
            idata += 1
            print(" {: .8e}".format(C_new[i,j]),end ="", file=f)
            if(idata%5==0 or idata==int(p.nbf*p.nbf)):
                print("",file=f)
#    print("Beta MO coefficients                       R   N=       {:6d}".format(p.nbf*p.nbf),file=f)
#    idata = 0
#    for j in range(p.nbf):
#        for i in range(p.nbf):
#            idata += 1
#            print(" {: .8e}".format(C_new[i,j]),end ="", file=f)
#            if(idata%5==0 or idata==int(p.nbf*p.nbf)):
#                print("",file=f)
    print("Total SCF Density                          R   N=       {:6d}".format(int(p.nbf*(p.nbf+1)/2)),file=f)
    DM = 2*np.einsum('i,mi,ni->mn',n,C_new[:,:p.nbf5],C_new[:,:p.nbf5])
    idata = 0
    for mu in range(p.nbf):
        for nu in range(mu+1):
            idata += 1
            print(" {: .8e}".format(DM[mu,nu]),end ="", file=f)
            if(idata%5==0 or idata==int(p.nbf*(p.nbf+1)/2)):
                print("",file=f)
    print("Natural orbital occupancies                R   N=       {:6d}".format(p.nbf5),file=f)
    for i in range(p.nbf5):
        print(" {: .8e}".format(n[i]),end ="", file=f)
        if((i+1)%5==0 or i==p.nbf5):
            print("",file=f)

    f.close()



from time import time
from scipy.linalg import eig,eigh
import psi4
import numpy as np
from numba import prange,njit
import pynof
try:
    import cupy as cp
except:
    pass

def compute_integrals(wfn,mol,p):

    # Integrador
    mints = psi4.core.MintsHelper(wfn.basisset())
    
    # Overlap, Kinetics, Potential
    S = np.asarray(mints.ao_overlap())
    T = np.asarray(mints.ao_kinetic())
    V = np.asarray(mints.ao_potential())
    H = T + V
    
    Dipole = np.asarray(mints.ao_dipole())

    I = []
    b_mnl = []
    if (not p.RI):
        # Integrales de Repulsión Electrónica, ERIs (mu nu | sigma lambda)
        I = np.asarray(mints.ao_eri())
    else:

        orb = wfn.basisset()
        aux = psi4.core.BasisSet.build(mol, "DF_BASIS_SCF", "", "JKFIT", orb.blend())
        zero_bas = psi4.core.BasisSet.zero_ao_basis_set()

        mnk = np.transpose(mints.ao_eri(aux, zero_bas, orb, orb))
        #Ppq = mints.ao_eri(orb, orb, aux, zero_bas)
        
        metric = mints.ao_eri(aux, zero_bas, aux, zero_bas)
        metric.power(-0.5, 1.e-14)
        p.nbfaux = metric.shape[0]

        mnk = np.squeeze(mnk)
        metric = np.squeeze(metric)

        b_mnl = np.einsum('mnk,kl->mnl', mnk, metric, optimize=True)

    if(p.gpu):
        I = cp.array(I)
        b_mnl = cp.array(b_mnl)

    return S,T,V,H,I,b_mnl,Dipole

def compute_geom_gradients(wfn,mol,n,C,cj12,ck12,elag,p):

    mints = psi4.core.MintsHelper(wfn.basisset())

    RDM1 = 2*np.einsum('p,mp,np->mn',n,C[:,:p.nbf5],C[:,:p.nbf5],optimize=True)
    lag = 2*np.einsum('mq,qp,np->mn',C,elag,C,optimize=True)
    
    grad = np.zeros((p.natoms,3))

    grad += np.array(mol.nuclear_repulsion_energy_deriv1())

    for i in range(p.natoms):
        dSx,dSy,dSz = np.array(mints.ao_oei_deriv1("OVERLAP",i))
        grad[i,0] -= np.einsum('mn,mn->',lag,dSx,optimize=True)
        grad[i,1] -= np.einsum('mn,mn->',lag,dSy,optimize=True)
        grad[i,2] -= np.einsum('mn,mn->',lag,dSz,optimize=True)

        dTx,dTy,dTz = np.array(mints.ao_oei_deriv1("KINETIC",i))
        grad[i,0] += np.einsum('mn,mn->',RDM1,dTx,optimize=True)
        grad[i,1] += np.einsum('mn,mn->',RDM1,dTy,optimize=True)
        grad[i,2] += np.einsum('mn,mn->',RDM1,dTz,optimize=True)

        dVx,dVy,dVz = np.array(mints.ao_oei_deriv1("POTENTIAL",i))
        grad[i,0] += np.einsum('mn,mn->',RDM1,dVx,optimize=True)
        grad[i,1] += np.einsum('mn,mn->',RDM1,dVy,optimize=True)
        grad[i,2] += np.einsum('mn,mn->',RDM1,dVz,optimize=True)

    np.fill_diagonal(cj12,0) # Remove diag.
    np.fill_diagonal(ck12,0) # Remove diag.

    if not p.RI:

        RDM2 = np.einsum('pq,mp,np,sq,lq->mnsl',cj12,C[:,:p.nbf5],C[:,:p.nbf5],C[:,:p.nbf5],C[:,:p.nbf5],optimize=True)
        RDM2 += np.einsum('p,mp,np,sp,lp->mnsl',n[:p.nbeta],C[:,:p.nbeta],C[:,:p.nbeta],C[:,:p.nbeta],C[:,:p.nbeta],optimize=True)
        RDM2 += np.einsum('p,mp,np,sp,lp->mnsl',n[p.nalpha:p.nbf5],C[:,p.nalpha:p.nbf5],C[:,p.nalpha:p.nbf5],C[:,p.nalpha:p.nbf5],C[:,p.nalpha:p.nbf5],optimize=True)
        RDM2 -= np.einsum('pq,mp,lp,sq,nq->mnsl',ck12,C[:,:p.nbf5],C[:,:p.nbf5],C[:,:p.nbf5],C[:,:p.nbf5],optimize=True)

        for i in range(p.natoms):
            derix,deriy,deriz = np.array(mints.ao_tei_deriv1(i))
            grad[i,0] += np.einsum("mnsl,mnsl->",RDM2,derix,optimize=True)
            grad[i,1] += np.einsum("mnsl,mnsl->",RDM2,deriy,optimize=True)
            grad[i,2] += np.einsum("mnsl,mnsl->",RDM2,deriz,optimize=True)

    else:

        orb = wfn.basisset()
        aux = psi4.core.BasisSet.build(mol, "DF_BASIS_SCF", "", "JKFIT", orb.blend())
        mints.set_basisset("aux",aux)
        zero_bas = psi4.core.BasisSet.zero_ao_basis_set()

        mnP = np.transpose(mints.ao_eri(aux, zero_bas, orb, orb))
        mnP = np.squeeze(mnP)

        metric = mints.ao_eri(aux, zero_bas, aux, zero_bas)
        metric.power(-1.0, 1.e-14)
        metric = np.squeeze(metric)

        tmp1 = np.einsum('mp,mnP->pnP',C[:,:p.nbf5],mnP,optimize=True)
        tmp1 = np.einsum('nq,pnP->pqP',C[:,:p.nbf5],tmp1,optimize=True)
        tmp1 = np.einsum('pqP,PQ->pqQ',tmp1,metric,optimize=True)
        tmp2 = np.einsum('pq,pqQ->pqQ',-ck12,tmp1,optimize=True)
        tmp3 = np.einsum('sq,pqQ->psQ',C[:,:p.nbf5],tmp2,optimize=True)
        val1 = np.einsum('lp,psQ->lsQ',C[:,:p.nbf5],tmp3,optimize=True)
        tmp3 = None
        val2 = np.einsum('pqQ,pqR->QR',tmp1,tmp2,optimize=True)

        tmp2 = np.einsum('pq,ppQ->qQ',cj12,tmp1,optimize=True)
        val1 += np.einsum('sq,lq,qQ->slQ',C[:,:p.nbf5],C[:,:p.nbf5],tmp2,optimize=True)
        val2 += np.einsum('qqQ,qR->QR',tmp1,tmp2,optimize=True)

        tmp2 = np.einsum('p,ppQ->pQ',n[:p.nbeta],tmp1[:p.nbeta,:p.nbeta],optimize=True)
        val1 += np.einsum('lp,sp,pQ->lsQ',C[:,:p.nbeta],C[:,:p.nbeta],tmp2,optimize=True)
        val2 += np.einsum('ppQ,pR->QR',tmp1[:p.nbeta,:p.nbeta],tmp2,optimize=True)

        tmp2 = np.einsum('p,ppQ->pQ',n[p.nalpha:p.nbf5],tmp1[p.nalpha:p.nbf5,p.nalpha:p.nbf5],optimize=True)
        val1 += np.einsum('lp,sp,pQ->lsQ',C[:,p.nalpha:p.nbf5],C[:,p.nalpha:p.nbf5],tmp2,optimize=True)
        val2 += np.einsum('ppQ,pR->QR',tmp1[p.nalpha:p.nbf5,p.nalpha:p.nbf5],tmp2,optimize=True)

        for i in range(p.natoms):
            d3x,d3y,d3z = np.array(mints.ao_3center_deriv1(i,"aux"))
            d2x,d2y,d2z = np.array(mints.ao_metric_deriv1(i,"aux"))
            grad[i,0] += 2*np.einsum('lsQ,Qsl->',val1,d3x,optimize=True)
            grad[i,0] -= np.einsum('QR,QR->',val2,d2x,optimize=True)
            grad[i,1] += 2*np.einsum('lsQ,Qsl->',val1,d3y,optimize=True)
            grad[i,1] -= np.einsum('QR,QR->',val2,d2y,optimize=True)
            grad[i,2] += 2*np.einsum('lsQ,Qsl->',val1,d3z,optimize=True)
            grad[i,2] -= np.einsum('QR,QR->',val2,d2z,optimize=True)

    return grad


######################################### J_mn^(j) K_mn^(j) #########################################

def computeJKj(C,I,b_mnl,p):

    if(p.gpu):
        if(p.RI):
            J,K = JKj_RI_GPU(C,b_mnl,p)
        else:
            J,K = JKj_Full_GPU(C,I,p)
    else:
        if(p.jit):
            if(p.RI):
                J,K = JKj_RI_jit(C,b_mnl,p.nbf,p.nbf5,p.nbfaux)
            else:
                J,K = JKj_Full_jit(C,I,p.nbf,p.nbf5)
        else:
            if(p.RI):
                J,K = JKj_RI(C,b_mnl,p)
            else:
                J,K = JKj_Full(C,I,p)

    return J,K

#########################################


def JKj_Full(C,I,p):
    #denmatj
    D = np.einsum('mi,ni->imn', C[:,0:p.nbf5], C[:,0:p.nbf5], optimize=True)
    #hstarj
    J = np.tensordot(D, I, axes=([1,2],[2,3]))
    #hstark
    K = np.tensordot(D, I, axes=([1,2],[1,3]))
    
    return J,K

def JKj_RI(C,b_mnl,p):
    
    #b_transform
    b_qnl = np.tensordot(C[:,0:p.nbf5],b_mnl, axes=([0],[0]))
    b_qql = np.einsum('nq,qnl->ql',C[:,0:p.nbf5],b_qnl, optimize=True)
    #hstarj
    J = np.tensordot(b_qql, b_mnl, axes=([1],[2]))
    #hstark
    K = np.einsum('qml,qnl->qmn', b_qnl, b_qnl, optimize=True)
    
    return J,K

def JKj_Full_GPU(C,I,p):

    #denmatj
    D = cp.einsum('mi,ni->imn', C[:,0:p.nbf5], C[:,0:p.nbf5], optimize=True)
    #hstarj
    J = cp.tensordot(D, I, axes=([1,2],[2,3]))
    #hstark
    K = np.einsum("inl,mnsl->ims",D,I,optimize=True)
    #K = cp.tensordot(D, I, axes=([1,2],[1,3]))
    
    return J.get(),K.get()

def JKj_RI_GPU(C,b_mnl,p):

    C = cp.array(C)
    #b_transform
    b_qnl = cp.tensordot(C[:,0:p.nbf5],b_mnl, axes=([0],[0]))
    b_qql = cp.einsum('nq,qnl->ql',C[:,0:p.nbf5],b_qnl, optimize=True)
    #hstarj
    J = cp.tensordot(b_qql, b_mnl, axes=([1],[2]))
    #hstark
    K = cp.einsum('qml,qnl->qmn', b_qnl, b_qnl, optimize=True)

    return J.get(),K.get()

@njit(parallel=True, cache=True)
def JKj_Full_jit(C,I,nbf,nbf5):

    #denmatj
    D = np.zeros((nbf5,nbf,nbf))
    for mu in prange(nbf):
        for nu in prange(mu+1):
            for i in prange(nbf5):
                D[i][mu][nu] = C[mu][i]*C[nu][i]
                D[i][nu][mu] = D[i][mu][nu]

    #hstarj
    J = np.zeros((nbf5,nbf,nbf))
    for i in prange(nbf5):
        for m in prange(nbf):
            for n in prange(m+1):
                for s in range(nbf):
                    for l in range(nbf):
                        J[i][m][n] += D[i][s][l]*I[m][n][s][l]
                J[i][n][m] = J[i][m][n]

    #hstark
    K = np.zeros((nbf5,nbf,nbf))
    for i in prange(nbf5):
        for m in prange(nbf):
            for s in prange(m+1):
                for n in range(nbf):
                    for l in range(nbf):
                        K[i][m][s] += D[i][n][l]*I[m][n][s][l]
                K[i][s][m] = K[i][m][s]

    return J,K

@njit(parallel=True, cache=True)
def JKj_RI_jit(C,b_mnl,nbf,nbf5,nbfaux):

    #denmatj
    J = np.zeros((nbf5,nbf,nbf))
    K = np.zeros((nbf5,nbf,nbf))
    for q in prange(nbf5):
        b_qnl = np.zeros((nbf,nbfaux))
        b_qql = np.zeros((nbfaux))
        for l in range(nbfaux):
            for n in range(nbf):
                for m in range(nbf):
                    b_qnl[n][l] += C[m][q]*b_mnl[m][n][l]
                b_qql[l] += C[n][q]*b_qnl[n][l]
            for n in range(nbf):
                for m in range(nbf):
                    K[q][m][n] += b_qnl[m][l]*b_qnl[n][l]
                    J[q][m][n] += b_qql[l]*b_mnl[m][n][l]

    return J,K

######################################### J_pq K_pq #########################################

def computeJKH_MO(C,H,I,b_mnl,p):

    if(p.gpu):
        if(p.RI):
            J_MO,K_MO,H_core = JKH_MO_RI_GPU(C,H,b_mnl,p)
        else:
            J_MO,K_MO,H_core = JKH_MO_Full_GPU(C,H,I,p)
    else:
        if(p.jit):
            if(p.RI):
                J_MO,K_MO,H_core = JKH_MO_RI_jit(C,H,b_mnl,p.nbf,p.nbf5,p.nbfaux)
            else:
                J_MO,K_MO,H_core = JKH_MO_Full_jit(C,H,I,p.nbf,p.nbf5)
        else:
            if(p.RI):
                J_MO,K_MO,H_core = JKH_MO_RI(C,H,b_mnl,p)
            else:
                J_MO,K_MO,H_core = JKH_MO_Full(C,H,I,p)

    return J_MO,K_MO,H_core

#########################################

def JKH_MO_Full(C,H,I,p):

    #denmatj
    D = np.einsum('mi,ni->imn', C[:,0:p.nbf5], C[:,0:p.nbf5],optimize=True)
    #QJMATm
    J = np.tensordot(D, I, axes=([1,2],[2,3]))
    J_MO = np.tensordot(J, D,axes=((1,2),(1,2)))
    #QKMATm
    K = np.tensordot(D, I, axes=([1,2],[1,3]))
    K_MO = np.tensordot(K, D, axes=([1,2],[1,2]))
    #QHMATm
    H_core = np.tensordot(D, H, axes=([1,2],[0,1]))

    return J_MO,K_MO,H_core

def JKH_MO_RI(C,H,b_mnl,p):

    #denmatj
    D = np.einsum('mi,ni->imn', C[:,0:p.nbf5], C[:,0:p.nbf5],optimize=True)
    #b transform
    b_pnl = np.tensordot(C[:,0:p.nbf5],b_mnl, axes=([0],[0]))
    b_pql = np.einsum('nq,pnl->pql',C[:,0:p.nbf5],b_pnl, optimize=True)
    #QJMATm
    J_MO = np.einsum('ppl,qql->pq', b_pql, b_pql, optimize=True)
    #QKMATm
    K_MO = np.einsum('pql,pql->pq', b_pql, b_pql, optimize=True)
    #QHMATm
    H_core = np.tensordot(D,H, axes=([1,2],[0,1]))

    return J_MO,K_MO,H_core

def JKH_MO_Full_GPU(C,H,I,p):

    C = cp.array(C)
    H = cp.array(H)
    #denmatj
    D = cp.einsum('mi,ni->imn', C[:,0:p.nbf5], C[:,0:p.nbf5],optimize=True)
    #QJMATm
    J = cp.tensordot(D, I, axes=([1,2],[2,3]))
    J_MO = cp.tensordot(J, D,axes=((1,2),(1,2)))
    #QKMATm
    K = cp.tensordot(D, I, axes=([1,2],[1,3]))
    K_MO = cp.tensordot(K, D, axes=([1,2],[1,2]))
    #QHMATm
    H_core = cp.tensordot(D, H, axes=([1,2],[0,1]))

    return J_MO.get(),K_MO.get(),H_core.get()

def JKH_MO_RI_GPU(C,H,b_mnl,p):

    C = cp.array(C)
    H = cp.array(H)
    #denmatj
    D = cp.einsum('mi,ni->imn', C[:,0:p.nbf5], C[:,0:p.nbf5],optimize=True)
    #b transform
    b_pnl = cp.tensordot(C[:,0:p.nbf5],b_mnl, axes=([0],[0]))
    b_pql = cp.einsum('nq,pnl->pql',C[:,0:p.nbf5],b_pnl, optimize=True)
    #QJMATm
    J_MO = cp.einsum('ppl,qql->pq', b_pql, b_pql, optimize=True)
    #QKMATm
    K_MO = cp.einsum('pql,pql->pq', b_pql, b_pql, optimize=True)
    #QHMATm
    H_core = cp.tensordot(D,H, axes=([1,2],[0,1]))

    return J_MO.get(),K_MO.get(),H_core.get()

@njit(parallel=True, cache=True)
def JKH_MO_Full_jit(C,H,I,nbf,nbf5):

    #denmatj
    D = np.zeros((nbf5,nbf,nbf))
    for mu in prange(nbf):
        for nu in prange(mu+1):
            for i in prange(nbf5):
                D[i][mu][nu] = C[mu][i]*C[nu][i]
                D[i][nu][mu] = D[i][mu][nu]
    #QJMATm
    J = np.zeros((nbf5,nbf,nbf))
    for i in prange(nbf5):
        for m in prange(nbf):
            for n in prange(m+1):
                for s in range(nbf):
                    for l in range(nbf):
                        J[i][m][n] += D[i][s][l]*I[m][n][s][l]
                J[i][n][m] = J[i][m][n]
    J_MO = np.zeros((nbf5,nbf5))
    for i in prange(nbf5):
        for j in prange(i+1):
            for m in range(nbf):
                for n in range(nbf):
                    J_MO[i][j] += D[j][m][n]*J[i][m][n]
            J_MO[j][i] = J_MO[i][j]

    #QKMATm
    K = np.zeros((nbf5,nbf,nbf))
    for i in prange(nbf5):
        for m in prange(nbf):
            for s in prange(m+1):
                for n in range(nbf):
                    for l in range(nbf):
                        K[i][m][s] += D[i][n][l]*I[m][n][s][l]
                K[i][s][m] = K[i][m][s]
    K_MO = np.zeros((nbf5,nbf5))
    for i in prange(nbf5):
        for j in prange(i+1):
            for m in range(nbf):
                for s in range(nbf):
                    K_MO[i][j] += D[j][m][s]*K[i][m][s]
            K_MO[j][i] = K_MO[i][j]

    #QHMATm
    H_core = np.zeros((nbf5))
    for i in prange(nbf5):
        for m in range(nbf):
            for n in range(m):
                H_core[i] += 2*D[i][m][n]*H[m][n]
            H_core[i] += D[i][m][m]*H[m][m]

    return J_MO,K_MO,H_core

@njit(parallel=True, cache=True)
def JKH_MO_RI_jit(C,H,b_mnl,nbf,nbf5,nbfaux):

    b_pql = np.zeros((nbf5,nbf5,nbfaux))
    for p in prange(nbf5):
        for n in range(nbf):
            for l in range(nbfaux):
                b_pnl = 0
                for m in range(nbf):
                    b_pnl += C[m][p]*b_mnl[m][n][l]
                for q in range(p+1):
                    b_pql[p][q][l] += C[n][q]*b_pnl

    H_core = np.zeros((nbf5))
    J_MO = np.zeros((nbf5,nbf5))
    K_MO = np.zeros((nbf5,nbf5))
    for p in prange(nbf5):
        for m in range(nbf):
            for n in range(nbf):
                H_core[p] += C[m][p]*H[m][n]*C[n][p]
        for q in range(p+1):
            for l in range(nbfaux):
                J_MO[p][q] += b_pql[p][p][l]*b_pql[q][q][l]
                K_MO[p][q] += b_pql[p][q][l]*b_pql[p][q][l]
            J_MO[q][p] = J_MO[p][q]
            K_MO[q][p] = K_MO[p][q]

    return J_MO,K_MO,H_core

######################################### J_mn^(j) K_mn^(j) #########################################

def computeD_HF(C,p):

    if(p.gpu):
        D = cp.einsum('mj,nj->mn',C[:,:p.nbeta],C[:,:p.nbeta],optimize=True)
        return D.get()
    else:
        D = np.einsum('mj,nj->mn',C[:,:p.nbeta],C[:,:p.nbeta],optimize=True)
        return D

def computeDalpha_HF(C,p):

    if(p.gpu):
        D = cp.einsum('mj,nj->mn',C[:,p.nbeta:p.nalpha],C[:,p.nbeta:p.nalpha],optimize=True)
        return D.get()
    else:
        D = np.einsum('mj,nj->mn',C[:,p.nbeta:p.nalpha],C[:,p.nbeta:p.nalpha],optimize=True)
        return D

def computeJK_HF(D,I,b_mnl,p):

    if(p.gpu):
#        if(p.RI):
#            J,K = JKj_RI_GPU(C,b_mnl,p)
#        else:
        J,K = JK_HF_Full_GPU(D,I,p)
    else:
#        if(p.RI):
#            J,K = JKj_RI_jit(C,b_mnl,p.nbf,p.nbf5,p.nbfaux)
#        else:
        J,K = JK_HF_Full_jit(D,I,p.nbeta,p.nbf,p.nbf5)

    return J,K

def JK_HF_Full_jit(D,I,nbeta,nbf,nbf5):

    J = np.einsum("ls,mnsl->mn",D,I,optimize=True)
    K = np.einsum("nl,mnsl->ms",D,I,optimize=True)

    return J,K

def JK_HF_Full_GPU(D,I,p):

    #denmatj
    J = cp.einsum("ls,mnsl->mn",D,I,optimize=True)
    K = cp.einsum("nl,mnsl->ms",D,I,optimize=True)

    return J.get(),K.get()

def computeJKalpha_HF(C,I,b_mnl,p):

    if(p.gpu):
#        if(p.RI):
#            J,K = JKj_RI_GPU(C,b_mnl,p)
#        else:
        D,J,K = JKalpha_HF_Full_GPU(C,I,p)
    else:
#        if(p.RI):
#            J,K = JKj_RI_jit(C,b_mnl,p.nbf,p.nbf5,p.nbfaux)
#        else:
        D,J,K = JKalpha_HF_Full_jit(C,I,p.nbeta,p.nalpha,p.nbf,p.nbf5)

    return D,J,K

@njit(parallel=True, cache=True)
def JKalpha_HF_Full_jit(C,I,nbeta,nalpha,nbf,nbf5):

    #denmatj
    D = np.zeros((nbf,nbf))
    for mu in prange(nbf):
        for nu in prange(mu):
            for i in prange(nbeta,nalpha):
                D[mu][nu] += C[mu][i]*C[nu][i]
            D[nu][mu] = D[mu][nu]

    #hstarj
    J = np.zeros((nbf,nbf))
    for m in prange(nbf):
        for n in prange(m):
            for s in range(nbf):
                for l in range(nbf):
                    J[m][n] += D[s][l]*I[m][n][s][l]
            J[n][m] = J[m][n]

    #hstark
    K = np.zeros((nbf,nbf))
    for m in prange(nbf):
        for s in prange(m):
            for n in range(nbf):
                for l in range(nbf):
                    K[m][s] += D[n][l]*I[m][n][s][l]
            K[s][m] = K[m][s]

    return D,J,K

def JKalpha_HF_Full_GPU(C,I,p):

    #denmatj
    D = cp.einsum("mj,nj->mn",C[:,p.nbeta:p.nalpha],C[:,p.nbeta:p.nalpha],optimize=True)
    J = cp.einsum("ls,mnsl->mn",D,I,optimize=True)
    K = cp.einsum("nl,mnsl->ms",D,I,optimize=True)

    return D.get(),J.get(),K.get()

def compute_iajb(C,I,b_mnl,p):

    if(p.gpu):
        iajb = iajb_Full_GPU(C,I,p)
    else:
        iajb = iajb_Full_jit(C,I,p.no1,p.nalpha,p.nbf,p.nbf5)

    return iajb

def iajb_Full_jit(C,I,no1,nalpha,nbf,nbf5):

    iajb = np.einsum('mi,na,mnsl,sj,lb->iajb',C[:,no1:nalpha],C[:,nalpha:nbf],I,C[:,no1:nalpha],C[:,nalpha:nbf],optimize=True)

    return iajb

def iajb_Full_GPU(C,I,p):

    iajb = cp.einsum('mi,na,mnsl,sj,lb->iajb',C[:,p.no1:p.nalpha],C[:,p.nalpha:p.nbf],I,C[:,p.no1:p.nalpha],C[:,p.nalpha:p.nbf],optimize=True)

    return iajb.get()

def compute_pqrt(C,I,b_mnl,p):

    if(p.gpu):
        pqrt = pqrt_Full_GPU(C,I,p)
    else:
        pqrt = pqrt_Full_jit(C,I,p.no1,p.nalpha,p.nbf,p.nbf5)

    return pqrt


def pqrt_Full_jit(C,I,no1,nalpha,nbf,nbf5):

    pqrt = np.einsum('mp,nq,mnsl,sr,lt->pqrt',C,C,I,C,C,optimize=True)

    return pqrt

def pqrt_Full_GPU(C,I,p):

    pqrt = cp.einsum('mp,nq,mnsl,sr,lt->pqrt',C,C,I,C,C,optimize=True)

    return pqrt.get()

def JKH_MO_tmp(C,H,I,b_mnl,p):

    if(p.gpu):
        if(p.RI):
            H_MO, b_MO = Integrals_MO_RI_GPU(C,H,b_mnl,p)
            pass
        else:
            H_MO, I_MO = Integrals_MO_Full_GPU(C,H,I,p)
    else:
        if(p.RI):
            H_MO, b_MO = Integrals_MO_RI_CPU(C,H,b_mnl,p)
        else:
            H_MO, I_MO = Integrals_MO_Full_CPU(C,H,I,p)

    if(p.RI):
        return H_MO,b_MO
    else:
        return H_MO,I_MO

def Integrals_MO_Full_CPU(C,H,I,p):

    H_mat = np.einsum("mi,mn,nj->ij",C,H,C[:,:p.nbf5],optimize=True)
    I_MO = np.einsum("mp,nq,mnsl,sr,lt->pqrt",C,C[:,:p.nbf5],I,C[:,:p.nbf5],C[:,:p.nbf5],optimize=True)

    return H_mat,I_MO

def Integrals_MO_Full_GPU(C,H,I,p):

    H_mat = cp.einsum("mi,mn,nj->ij",C,H,C[:,:p.nbf5],optimize=True)
    I_MO = cp.einsum("mp,nq,mnsl,sr,lt->pqrt",C,C[:,:p.nbf5],I,C[:,:p.nbf5],C[:,:p.nbf5],optimize=True)

    return H_mat,I_MO

def Integrals_MO_RI_CPU(C,H,b_mnl,p):

    H_mat = np.einsum("mi,mn,nj->ij",C,H,C[:,:p.nbf5],optimize=True)
    b_MO = np.einsum("mp,nq,mnl->pql",C,C[:,:p.nbf5],b_mnl,optimize=True)

    return H_mat,b_MO

def Integrals_MO_RI_GPU(C,H,b_mnl,p):

    H_mat = cp.einsum("mi,mn,nj->ij",C,H,C[:,:p.nbf5],optimize=True)
    b_MO = cp.einsum("mp,nq,mnl->pql",C,C[:,:p.nbf5],b_mnl,optimize=True)

    return H_mat,b_MO




from .input import *
from .energy import *
from .parameters import *
from .integrals import *
from .guess import *
from .minimization import *
from .optimization import *
from .pnof import *
from .postpnof import *
from .utils import *
from .output import *



import numpy as np
from scipy.optimize import minimize
from time import time
import pynof 

def hfidr(C,H,I,b_mnl,E_nuc,p,printmode):

    no1_ori = p.no1
    p.no1 = p.nbeta

    n = np.zeros((p.nbf5))
    n[0:p.nbeta] = 1.0
    n[p.nbeta:p.nalpha] = 0.5

    cj12 = 2*np.einsum('i,j->ij',n,n)
    ck12 = np.einsum('i,j->ij',n,n)
    if(p.MSpin==0 and p.nsoc>1):
        ck12[p.nbeta:p.nalpha,p.nbeta:p.nalpha] = 2*np.einsum('i,j->ij',n[p.nbeta:p.nalpha],n[p.nbeta:p.nalpha])

    if(printmode):
        print("Hartree-Fock")
        print("============")
        print("")

        print('{:^7} {:^7} {:^14} {:^14} {:^15} {:^14}'.format("Nitext","Nitint","Eelec","Etot","Ediff","maxdiff"))

    E,elag,sumdiff,maxdiff = pynof.ENERGY1r(C,n,H,I,b_mnl,cj12,ck12,p)

    fmiug0 = None

    ext = True
    # iteraciones externas
    for i_ext in range(p.maxitid):
        if i_ext==0:
            maxlp = 1
        else:
            maxlp = p.maxloop

        # iteraciones internas
        for i_int in range(maxlp):
            E_old = E

            if(p.scaling):
                fmiug = pynof.fmiug_scaling(fmiug0,elag,i_ext,p.nzeros,p.nbf,p.noptorb)

            fmiug0, W = np.linalg.eigh(fmiug)
            C = np.matmul(C,W)
            E,elag,sumdiff,maxdiff = pynof.ENERGY1r(C,n,H,I,b_mnl,cj12,ck12,p)

            E_diff = E-E_old
            if(abs(E_diff)<p.thresheid):
                if(printmode):
                    print('{:6d} {:6d} {:14.8f} {:14.8f} {:14.8f} {:14.8f}'.format(i_ext,i_int,E,E+E_nuc,E_diff,maxdiff))
                for i in range(p.nbf):
                    fmiug0[i] = elag[i][i]
                ext = False
                break

        if(not ext):
            break
        if(printmode):
            print('{:6d} {:6d} {:14.8f} {:14.8f} {:14.8f} {:14.8f}'.format(i_ext,i_int,E,E+E_nuc,E_diff,maxdiff))


    # Regresamos no1 a su estado original
    p.no1 = no1_ori

    return E,C,fmiug0

def occoptr(gamma,C,H,I,b_mnl,p):

    J_MO,K_MO,H_core = pynof.computeJKH_MO(C,H,I,b_mnl,p)

    E = 0
    nit = 0
    success = True

    pynof.calcocce(gamma,J_MO,K_MO,H_core,p)

    if (p.ndoc>0):
        res = minimize(pynof.calcocce, gamma, args=(J_MO,K_MO,H_core,p), jac=pynof.calcoccg, method=p.occupation_optimizer)
        gamma = res.x
        E = res.fun
        nit = res.nit
        success = res.success

    n,dR = pynof.ocupacion(gamma,p.no1,p.ndoc,p.nalpha,p.nv,p.nbf5,p.ndns,p.ncwo,p.HighSpin,p.occ_method)
    cj12,ck12 = pynof.PNOFi_selector(n,p)

    return E,nit,success,gamma,n,cj12,ck12

def orboptr(C,n,H,I,b_mnl,cj12,ck12,i_ext,itlim,fmiug0,p,printmode):

    i_int = 0
    success_orb = False

    E,elag,sumdiff,maxdiff = pynof.ENERGY1r(C,n,H,I,b_mnl,cj12,ck12,p)

    if(maxdiff<p.threshl):# and abs(E_diff)<p.threshe):
        success_orb = True
        return E,C,i_int,success_orb,itlim,fmiug0

    if (p.scaling and i_ext>1 and i_ext >= itlim):# and sumdiff > sumdiff_old):
        p.nzeros = p.nzeros + 1
        itlim = i_ext + p.itziter
        if (p.nzeros>p.nzerosm):
            p.nzeros = p.nzerosr
        #if (p.nzeros>abs(int(np.log10(maxdiff)))+1):
        #    p.nzeros = p.nzerosr
            #p.nzeros = abs(int(np.log10(maxdiff)))
    sumdiff_old = sumdiff

    if i_ext==0:
        maxlp = 1
    else:
        maxlp = p.maxloop

    fmiug = np.zeros((p.noptorb,p.noptorb))
    fk = np.zeros((30,p.noptorb,p.noptorb))
    bdiis = np.zeros((31,31))
    cdiis = np.zeros((31))
    iloop = 0
    idiis = 0

    for i_int in range(maxlp):
        iloop = iloop + 1
        E_old2 = E
        
        #scaling
        if(p.scaling):
            fmiug = pynof.fmiug_scaling(fmiug0,elag,i_ext,p.nzeros,p.nbf,p.noptorb)
        if(p.diis and maxdiff < p.thdiis):
            fk,fmiug,idiis,bdiis = pynof.fmiug_diis(fk,fmiug,idiis,bdiis,cdiis,maxdiff,p.noptorb,p.ndiis,p.perdiis)

        eigval, eigvec = np.linalg.eigh(fmiug)
        fmiug0 = eigval

        C = np.matmul(C,eigvec)

        E,elag,sumdiff,maxdiff = pynof.ENERGY1r(C,n,H,I,b_mnl,cj12,ck12,p)

        E_diff2 = E-E_old2

        if(abs(E_diff2)<p.threshec or i_int==maxlp-1):
            break

    return E,C,i_int,success_orb,itlim,fmiug0

def orbopt_rotations(gamma,C,H,I,b_mnl,p):

    y = np.zeros((p.nvar))

    n,dn_dgamma = pynof.ocupacion(gamma,p.no1,p.ndoc,p.nalpha,p.nv,p.nbf5,p.ndns,p.ncwo,p.HighSpin,p.occ_method)
    cj12,ck12 = pynof.PNOFi_selector(n,p)

    if("trust" in p.orbital_optimizer or "Newton-CG" in p.orbital_optimizer):
        res = minimize(pynof.calcorbeg, y, args=(n,cj12,ck12,C,H,I,b_mnl,p),jac=True,hess="2-point",method=p.orbital_optimizer,options={"maxiter":p.maxloop})
    else:
        res = minimize(pynof.calcorbeg, y, args=(n,cj12,ck12,C,H,I,b_mnl,p),jac=True,method=p.orbital_optimizer,options={"maxiter":p.maxloop})

    E = res.fun
    y = res.x
    grad = res.jac
    nit = res.nit
    success = res.success

    C = pynof.rotate_orbital(y,C,p)

    return E,C,nit,success

def comb(gamma,C,H,I,b_mnl,p):

    x = np.zeros((p.nvar+p.nv))
    x[p.nvar:] = gamma

    if("trust" in p.orbital_optimizer or "Newton-CG" in p.orbital_optimizer):
        res = minimize(pynof.calccombeg, x, args=(C,H,I,b_mnl,p),jac=True,hess="2-point",method=p.combined_optimizer,options={"maxiter":p.maxloop})
    else:
        res = minimize(pynof.calccombeg, x, args=(C,H,I,b_mnl,p),jac=True,method=p.combined_optimizer,options={"maxiter":p.maxloop})

    E = res.fun
    x = res.x
    grad = res.jac
    nit = res.nit
    y = x[:p.nvar]
    gamma = x[p.nvar:]
    C = pynof.rotate_orbital(y,C,p)

    n,dR = pynof.ocupacion(gamma,p.no1,p.ndoc,p.nalpha,p.nv,p.nbf5,p.ndns,p.ncwo,p.HighSpin,p.occ_method)

    return E,C,gamma,n,grad,res.nit,res.success
